from shapely.geometry import Polygon, Point
from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS, WGS84_003
from vgrid.utils.rhealpixdggs.cell import Cell
from itertools import compress


def get_finest_containing_cell(
    polygon: Polygon, rdggs: RHEALPixDGGS = WGS84_003
) -> Cell:
    """
    Finds the finest DGGS Cell containing a given cartesian polygon
    """

    def _get_finest_cell(polygon, suid, rdggs=rdggs):
        parent_cell = Cell(rdggs=rdggs, suid=suid)
        # get the children cells and polygons for these cells
        children_cells = [cell for cell in parent_cell.subcells()]
        children_poly = [Polygon(cell.vertices(plane=False)) for cell in children_cells]
        # function and truth list for multipolygon / polygon (polygon) contained within multipolygon / polygon (cell)
        truth = [poly.contains(polygon) for poly in children_poly]
        # if we get something back, check the next level lower
        returned_cells = list(compress(children_cells, truth))
        if returned_cells:
            finest = _get_finest_cell(polygon, returned_cells[0].suid, rdggs)
        else:
            parent_poly = Polygon(parent_cell.vertices(plane=False))
            if parent_poly.contains(polygon):
                finest = parent_cell
            else:
                finest = None
        return finest

    for suid in [tuple(x) for x in ["N", "O", "P", "Q", "R", "S"]]:
        finest = _get_finest_cell(polygon, suid, rdggs)
        if finest is not None:
            return finest


# TODO class should be a general class for collections of cells (believe the term is 'zone'?)
class CellZoneFromPoly:
    def __init__(
        self,
        feature,
        res_limit,
        return_cells: bool,
        file=None,
        bounding_cell=None,
        rdggs=WGS84_003,
    ):
        self.label = feature[0]
        self.geometry = feature[1]
        self.res_limit = res_limit
        self.return_cells = return_cells
        if return_cells:
            self.cells_list = []
        # self.i = 0
        self.file = file
        if file:
            self.file.write(f"\n{self.label},")
        if bounding_cell is None:
            self._get_dggs_poly(get_finest_containing_cell(self.geometry, rdggs))
        else:
            self._get_dggs_poly(bounding_cell)

    def _get_dggs_poly(self, bounding_cell):
        bounding_poly = Polygon(bounding_cell.vertices(plane=False))
        if self.geometry.contains(
            bounding_poly
        ):  # edge case where the polygon is the same as the bounding cell
            self._write_cells(bounding_cell, bounding_poly, "bounding poly")
        else:
            if bounding_cell.resolution + 1 > self.res_limit:
                pass
            else:
                children_cells = [cell for cell in bounding_cell.subcells()]
                children_poly = [
                    Polygon(cell.vertices(plane=False)) for cell in children_cells
                ]
                together = list(zip(children_cells, children_poly))
                # print(f'processing chhildren for bounding cell {bounding_cell}')
                self._process_children(together)
        # print(f'*bounding cell* {bounding_cell}')
        # for i in self.cells_list:
        #     print(i)
        # print('****************')
        return self.cells_list

    def _process_children(self, together):
        for child_cell, child_poly in together:
            # 1: add contained cells
            if self.geometry.contains(child_poly):
                self._write_cells(child_cell, child_poly, "fully contained")
            # 2: check we're not at the limit, if we are, check centroids
            elif child_cell.resolution == self.res_limit:
                if self.geometry.contains(Point(child_cell.nucleus(plane=False))):
                    self._write_cells(child_cell, child_poly, "nucleus")
            # 3: check the children (call this same function on the children)
            else:
                if self.geometry.overlaps(child_poly):
                    self._get_dggs_poly(child_cell)

    def _write_cells(self, cell, poly, desc):
        """
        Writes cell / polygon details to either or both of a list or a file.
        """
        if self.file is not None:
            self.file.write(f"{str(cell)} {desc}\n ")
            # self.file.write(f"{poly.wkt}\n")
        if self.return_cells:
            # cell = self.add_int_suid(cell)
            self.cells_list.append(cell)


def compress_order_cells(cells: list[str]) -> list[str]:
    """
    Compresses and sorts a set of cells
    """
    import re

    def alphanum_sort(lst: list[str]) -> list[str]:
        def convert(text):
            return int(text) if text.isdigit() else text
        def alphanum_key(key):
            return [convert(c) for c in re.split("([0-9]+)", key)]
        return sorted(lst, key=alphanum_key)

    cells = set(cells)
    upper_cells = {}
    for cell in cells:
        upper_cells.setdefault(cell[:-1], []).append(cell)
    compressed_cells = []
    for k, v in upper_cells.items():
        if len(v) == 9:
            compressed_cells.append(k)
        else:
            compressed_cells.extend(v)
    return alphanum_sort(compressed_cells)



from shapely.geometry import box
import argparse
import platform

if platform.system() == "Linux":
    from vgrid.utils.dggrid4py import DGGRIDv7, dggs_types
    from vgrid.utils.dggrid4py.dggrid_runner import output_address_types


def generate_grid(dggrid_instance, dggs_type, resolution, bbox, address_type):
    def address_transform(dggrid_seqnum, dggs_type, resolution, address_type):
        address_type_transform = dggrid_instance.address_transform(
            [dggrid_seqnum],
            dggs_type=dggs_type,
            resolution=resolution,
            mixed_aperture_level=None,
            input_address_type="SEQNUM",
            output_address_type=address_type,
        )
        return address_type_transform.loc[0, address_type]

    if bbox:
        bounding_box = box(*bbox)
        dggrid_gdf = dggrid_instance.grid_cell_polygons_for_extent(
            dggs_type,
            resolution,
            clip_geom=bounding_box,
            split_dateline=True,
            output_address_type=address_type,
        )
        try:
            if address_type != "SEQNUM":
                dggrid_gdf["name"] = dggrid_gdf["name"].astype(str)
                dggrid_gdf["name"] = dggrid_gdf["name"].apply(
                    lambda val: address_transform(
                        val, dggs_type, resolution, address_type
                    )
                )
                dggrid_gdf = dggrid_gdf.rename(columns={"name": address_type.lower()})
            else:
                dggrid_gdf = dggrid_gdf.rename(columns={"name": "seqnum"})
        except Exception:
            pass
        geojson_path = f"dggrid_{dggs_type}_{resolution}_{address_type}_bbox.geojson"
        dggrid_gdf.to_file(geojson_path, driver="GeoJSON")

    else:
        dggrid_gdf = dggrid_instance.grid_cell_polygons_for_extent(
            dggs_type, resolution, split_dateline=True, output_address_type=address_type
        )
        try:
            if address_type != "SEQNUM":
                dggrid_gdf["name"] = dggrid_gdf["name"].astype(str)
                dggrid_gdf["name"] = dggrid_gdf["name"].apply(
                    lambda val: address_transform(
                        val, dggs_type, resolution, address_type
                    )
                )
                dggrid_gdf = dggrid_gdf.rename(columns={"name": address_type.lower()})
            else:
                dggrid_gdf = dggrid_gdf.rename(columns={"name": "seqnum"})
        except Exception:
            pass

        geojson_path = f"dggrid_{dggs_type}_{resolution}_{address_type}.geojson"
        dggrid_gdf.to_file(geojson_path, driver="GeoJSON")

    print(f"GeoJSON saved as {geojson_path}")


def main():
    if platform.system() == "Linux":
        parser = argparse.ArgumentParser(description="Generate DGGRID.")
        parser.add_argument(
            "-t",
            "--dggs_type",
            choices=dggs_types,
            help="Select a DGGS type from the available options.",
        )
        parser.add_argument(
            "-r", "--resolution", type=int, required=True, help="resolution"
        )
        parser.add_argument(
            "-b",
            "--bbox",
            type=float,
            nargs=4,
            help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
        )
        parser.add_argument(
            "-a",
            "--address_type",
            choices=output_address_types,
            help="Select an output address type.",
        )
        args = parser.parse_args()

        dggrid_instance = DGGRIDv7(
            executable="/usr/local/bin/dggrid",
            working_dir=".",
            capture_logs=False,
            silent=True,
            tmp_geo_out_legacy=False,
            debug=False,
        )
        resolution = args.resolution
        dggs_type = args.dggs_type
        bbox = args.bbox
        address_type = args.address_type
        try:
            generate_grid(dggrid_instance, dggs_type, resolution, bbox, address_type)
        except Exception:
            print(
                "Please ensure that -a <address_type> are set appropriately, and there is an excutable DGGRID located at /usr/local/bin/dggrid. Please install DGGRID following instructions from https://github.com/sahrk/DGGRID/blob/master/INSTALL.md"
            )
    else:
        print(
            "dggrid only works on Linux with an excutable DGGRID at /usr/local/bin/dggrid. Please install DGGRID following instructions from https://github.com/sahrk/DGGRID/blob/master/INSTALL.md"
        )


if __name__ == "__main__":
    main()



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import unittest
from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
from vgrid.utils.eaggr.shapes.dggs_linestring import DggsLinestring


## Unit tests for the DggsLinestring class.
class TestDggsLinestring(unittest.TestCase):
    def test_getters(self):
        orig_cells = [DggsCell("0000"), DggsCell("1111")]
        linestring = DggsLinestring()
        linestring.add_cell(orig_cells[0])
        linestring.add_cell(orig_cells[1])
        cells = linestring.get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0], orig_cells[0])
        self.assertEqual(cells[1], orig_cells[1])

    def test_errors(self):
        linestring = DggsLinestring()
        with self.assertRaises(ValueError):
            linestring.add_cell("Not a DggsCell")



import os
import argparse
import json
import re
import pandas as pd
import h3
from urllib.parse import urlparse

from shapely.geometry import Polygon, mapping
from vgrid.generator.h3grid import fix_h3_antimeridian_cells
from vgrid.generator.settings import geodesic_dggs_to_feature

from vgrid.utils import s2, olc, geohash, georef, mgrs, mercantile, maidenhead
from vgrid.utils.gars import garsgrid

from shapely.wkt import loads

from vgrid.utils.antimeridian import fix_polygon

from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS
from vgrid.utils.rhealpixdggs.ellipsoids import WGS84_ELLIPSOID
from vgrid.conversion.dggs2geojson import rhealpix_cell_to_polygon

import platform

if platform.system() == "Windows":
    from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
    from vgrid.utils.eaggr.enums.model import Model
    from vgrid.generator.isea4tgrid import fix_isea4t_wkt, fix_isea4t_antimeridian_cells

    from vgrid.conversion.dggs2geojson import isea3h_cell_to_polygon

from vgrid.utils.easedggs.constants import levels_specs
from vgrid.utils.easedggs.dggs.grid_addressing import grid_ids_to_geos

from vgrid.generator.settings import (
    isea3h_accuracy_res_dict,
    graticule_dggs_to_feature,
)
from vgrid.utils.qtm import constructGeometry, qtm_id_to_facet


from pyproj import Geod

geod = Geod(ellps="WGS84")
E = WGS84_ELLIPSOID


#################################################################################
#  H3
#################################################################################
def h32feature(h3_id):
    """Convert H3 cell ID to a GeoJSON Polygon."""
    cell_boundary = h3.cell_to_boundary(h3_id)
    if cell_boundary:
        filtered_boundary = fix_h3_antimeridian_cells(cell_boundary)
        # Reverse lat/lon to lon/lat for GeoJSON compatibility
        reversed_boundary = [(lon, lat) for lat, lon in filtered_boundary]
        cell_polygon = Polygon(reversed_boundary)
        resolution = h3.get_resolution(h3_id)
        num_edges = 6
        if h3.is_pentagon(h3_id):
            num_edges = 6
        h3_feature = geodesic_dggs_to_feature(
            "h3", h3_id, resolution, cell_polygon, num_edges
        )

        return h3_feature


def csv2h3(csv_file, column="h3"):
    """Convert CSV file containing H3 cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None
    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None
    geojson_features = []
    for _, row in df.iterrows():
        try:
            h3_id = row[column]
            h3_feature = h32feature(h3_id)
            if h3_feature:
                geojson_features.append(h3_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue
    if not geojson_features:
        print("No valid H3 cells found in the CSV file.")
        return None
    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2h3_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with H3 column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with 'h3' column")
    parser.add_argument(
        "-id", dest="id", help="Name of the H3 column (default: 'h3')", default="h3"
    )
    args = parser.parse_args()
    h3_csv = args.csv
    h3_id = args.id
    h3_geojson = csv2h3(h3_csv, h3_id)
    geojson_name = os.path.splitext(os.path.basename(h3_csv))[0]
    geojson_path = f"{geojson_name}2h3.geojson"
    with open(geojson_path, "w") as f:
        json.dump(h3_geojson, f)
    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  S2
#################################################################################
def s22feature(s2_token):
    # Create an S2 cell from the given cell ID
    cell_id = s2.CellId.from_token(s2_token)
    cell = s2.Cell(cell_id)
    if cell:
        # Get the vertices of the cell (4 vertices for a rectangular cell)
        vertices = [cell.get_vertex(i) for i in range(4)]
        # Prepare vertices in (longitude, latitude) format for Shapely
        shapely_vertices = []
        for vertex in vertices:
            lat_lng = s2.LatLng.from_point(vertex)  # Convert Point to LatLng
            longitude = lat_lng.lng().degrees  # Access longitude in degrees
            latitude = lat_lng.lat().degrees  # Access latitude in degrees
            shapely_vertices.append((longitude, latitude))

        # Close the polygon by adding the first vertex again
        shapely_vertices.append(shapely_vertices[0])  # Closing the polygon
        # Create a Shapely Polygon
        cell_polygon = fix_polygon(Polygon(shapely_vertices))  # Fix antimeridian
        resolution = cell_id.level()
        num_edges = 4

        s2_feature = geodesic_dggs_to_feature(
            "s2", s2_token, resolution, cell_polygon, num_edges
        )
        return s2_feature


def csv2s2(csv_file, column="s2"):
    """Convert CSV file containing S2 cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None
    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None
    geojson_features = []
    for _, row in df.iterrows():
        try:
            s2_id = row[column]
            s2_feature = s22feature(s2_id)
            if s2_feature:
                geojson_features.append(s2_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue
    if not geojson_features:
        print("No valid S2 cells found in the CSV file.")
        return None
    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2s2_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with S2 column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with S2 column")
    parser.add_argument(
        "-id", dest="id", help="Name of the S2 column (default: 's2')", default="s2"
    )
    args = parser.parse_args()
    s2_csv = args.csv
    s2_id = args.id
    s2_geojson = csv2s2(s2_csv, s2_id)
    geojson_name = os.path.splitext(os.path.basename(s2_csv))[0]
    geojson_path = f"{geojson_name}2s2.geojson"
    with open(geojson_path, "w") as f:
        json.dump(s2_geojson, f)
    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Rhealpix
#################################################################################
def rhealpix2feature(rhealpix_id):
    rhealpix_uids = (rhealpix_id[0],) + tuple(map(int, rhealpix_id[1:]))
    rhealpix_dggs = RHEALPixDGGS(ellipsoid=E, north_square=1, south_square=3, N_side=3)
    rhealpix_cell = rhealpix_dggs.cell(rhealpix_uids)
    if rhealpix_cell:
        resolution = rhealpix_cell.resolution
        cell_polygon = rhealpix_cell_to_polygon(rhealpix_cell)
        num_edges = 4
        if rhealpix_cell.ellipsoidal_shape() == "dart":
            num_edges = 3
        rhealpix_feature = geodesic_dggs_to_feature(
            "rhealpix", rhealpix_id, resolution, cell_polygon, num_edges
        )
        return rhealpix_feature


def csv2rhealpix(csv_file, column="rhealpix"):
    """Convert CSV file containing rHEALPix cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            rhealpix_id = row[column]
            rhealpix_feature = rhealpix2feature(rhealpix_id)
            if rhealpix_feature:
                geojson_features.append(rhealpix_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid rHEALPix cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2rhealpix_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with rhealpix column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with rhealpix column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the rhealpix column (default: 'rhealpix')",
        default="rhealpix",
    )
    args = parser.parse_args()
    rhealpix_csv = args.csv
    rhealpix_id = args.id

    rhealpix_geojson = csv2rhealpix(rhealpix_csv, rhealpix_id)
    geojson_name = os.path.splitext(os.path.basename(rhealpix_csv))[0]
    geojson_path = f"{geojson_name}2rhealpix.geojson"

    with open(geojson_path, "w") as f:
        json.dump(rhealpix_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Open-Eaggr ISEA4T
#################################################################################
def isea4t2feature(isea4t_id):
    if platform.system() == "Windows":
        isea4t_dggs = Eaggr(Model.ISEA4T)
        cell_to_shape = isea4t_dggs.convert_dggs_cell_outline_to_shape_string(
            DggsCell(isea4t_id), ShapeStringFormat.WKT
        )
        cell_to_shape_fixed = loads(fix_isea4t_wkt(cell_to_shape))
        if (
            isea4t_id.startswith("00")
            or isea4t_id.startswith("09")
            or isea4t_id.startswith("14")
            or isea4t_id.startswith("04")
            or isea4t_id.startswith("19")
        ):
            cell_to_shape_fixed = fix_isea4t_antimeridian_cells(cell_to_shape_fixed)

        if cell_to_shape_fixed:
            resolution = len(isea4t_id) - 2
            num_edges = 3
            cell_polygon = Polygon(list(cell_to_shape_fixed.exterior.coords))
            isea4t_feature = geodesic_dggs_to_feature(
                "isea4t", isea4t_id, resolution, cell_polygon, num_edges
            )
            return isea4t_feature


def csv2isea4t(csv_file, column="isea4t"):
    """Convert CSV file containing ISEA4T cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            isea4t_id = row[column]
            isea4t_feature = isea4t2feature(isea4t_id)
            if isea4t_feature:
                geojson_features.append(isea4t_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid ISEA4T cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2isea4t_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with ISEA4T column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with ISEA4T column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the ISEA4T column (default: 'isea4t')",
        default="isea4t",
    )
    args = parser.parse_args()
    isea4t_csv = args.csv
    isea4t_id = args.id

    isea4t_geojson = csv2isea4t(isea4t_csv, isea4t_id)
    geojson_name = os.path.splitext(os.path.basename(isea4t_csv))[0]
    geojson_path = f"{geojson_name}2isea4t.geojson"

    with open(geojson_path, "w") as f:
        json.dump(isea4t_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Open-Eaggr ISEA3H
#################################################################################
def isea3h2feature(isea3h_id):
    if platform.system() == "Windows":
        isea3h_dggs = Eaggr(Model.ISEA3H)
        cell_polygon = isea3h_cell_to_polygon(isea3h_id)
        if cell_polygon:
            cell_centroid = cell_polygon.centroid
            center_lat = round(cell_centroid.y, 7)
            center_lon = round(cell_centroid.x, 7)

            cell_area = round(abs(geod.geometry_area_perimeter(cell_polygon)[0]), 3)
            cell_perimeter = abs(geod.geometry_area_perimeter(cell_polygon)[1])

            isea3h2point = isea3h_dggs.convert_dggs_cell_to_point(DggsCell(isea3h_id))
            accuracy = isea3h2point._accuracy

            avg_edge_len = cell_perimeter / 6
            resolution = isea3h_accuracy_res_dict.get(accuracy)

            if resolution == 0:  # icosahedron faces at resolution = 0
                avg_edge_len = cell_perimeter / 3

            if accuracy == 0.0:
                if round(avg_edge_len, 2) == 0.06:
                    resolution = 33
                elif round(avg_edge_len, 2) == 0.03:
                    resolution = 34
                elif round(avg_edge_len, 2) == 0.02:
                    resolution = 35
                elif round(avg_edge_len, 2) == 0.01:
                    resolution = 36

                elif round(avg_edge_len, 3) == 0.007:
                    resolution = 37
                elif round(avg_edge_len, 3) == 0.004:
                    resolution = 38
                elif round(avg_edge_len, 3) == 0.002:
                    resolution = 39
                elif round(avg_edge_len, 3) <= 0.001:
                    resolution = 40

            isea3h_feature = {
                "type": "Feature",
                "geometry": mapping(cell_polygon),
                "properties": {
                    "isea3h": isea3h_id,
                    "resolution": resolution,
                    "center_lat": center_lat,
                    "center_lon": center_lon,
                    "avg_edge_len": round(avg_edge_len, 3),
                    "cell_area": cell_area,
                },
            }
            return isea3h_feature


def csv2isea3h(csv_file, id_col=None):
    if not os.path.exists(csv_file):
        print(f"Error: Input file {csv_file} does not exist.")
        return

    if id_col is None:
        id_col = "isea3h"

    try:
        df = pd.read_csv(csv_file, dtype=str)  # Read entire file

        if id_col not in df.columns:
            print(
                f"Error: Column '{id_col}' is missing in the input CSV. Please check and try again."
            )
            return

    except Exception as e:
        print(f"Error reading CSV file: {e}")
        return

    geojson_features = []
    for _, row in df.iterrows():
        try:
            isea3h_id = row[id_col]
            isea3h_feature = isea3h2feature(isea3h_id)
            if isea3h_feature:
                isea3h_feature["properties"].update(
                    row.to_dict()
                )  # Append all CSV data to properties
                geojson_features.append(isea3h_feature)
        except Exception as e:
            print(f" Skipping row {row.to_dict()}: {e}")

    isea3h_geojson = {"type": "FeatureCollection", "features": geojson_features}
    return isea3h_geojson


def csv2isea3h_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with ISEA3H column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with ISEA3H column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the ISEA3H column (default: 'isea3h')",
        default="isea3h",
    )
    args = parser.parse_args()
    isea3h_csv = args.csv
    isea3h_id = args.id

    isea3h_geojson = csv2isea3h(isea3h_csv, isea3h_id)
    geojson_name = os.path.splitext(os.path.basename(isea3h_csv))[0]
    geojson_path = f"{geojson_name}2isea3h.geojson"

    with open(geojson_path, "w") as f:
        json.dump(isea3h_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  EASE-DGGS
#################################################################################
def ease2feature(ease_id):
    level = int(ease_id[1])  # Get the level (e.g., 'L0' -> 0)
    # Get level specs
    level_spec = levels_specs[level]
    n_row = level_spec["n_row"]
    n_col = level_spec["n_col"]

    geo = grid_ids_to_geos([ease_id])
    center_lon, center_lat = geo["result"]["data"][0]

    cell_min_lat = center_lat - (180 / (2 * n_row))
    cell_max_lat = center_lat + (180 / (2 * n_row))
    cell_min_lon = center_lon - (360 / (2 * n_col))
    cell_max_lon = center_lon + (360 / (2 * n_col))

    cell_polygon = Polygon(
        [
            [cell_min_lon, cell_min_lat],
            [cell_max_lon, cell_min_lat],
            [cell_max_lon, cell_max_lat],
            [cell_min_lon, cell_max_lat],
            [cell_min_lon, cell_min_lat],
        ]
    )

    if cell_polygon:
        resolution = level
        num_edges = 4
        ease_feature = geodesic_dggs_to_feature(
            "ease", ease_id, resolution, cell_polygon, num_edges
        )
        return ease_feature


def csv2ease(csv_file, column="ease"):
    """Convert CSV file containing EASE cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            ease_id = row[column]
            ease_feature = ease2feature(ease_id)
            if ease_feature:
                geojson_features.append(ease_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid EASE cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2ease_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with EASE column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with EASE column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the EASE column (default: 'ease')",
        default="ease",
    )
    args = parser.parse_args()
    ease_csv = args.csv
    ease_id = args.id

    ease_geojson = csv2ease(ease_csv, ease_id)
    geojson_name = os.path.splitext(os.path.basename(ease_csv))[0]
    geojson_path = f"{geojson_name}2ease.geojson"

    with open(geojson_path, "w") as f:
        json.dump(ease_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  QTM
#################################################################################
def qtm2feature(qtm_id):
    qtm_id = str(qtm_id)
    facet = qtm_id_to_facet(qtm_id)
    cell_polygon = constructGeometry(facet)
    if cell_polygon:
        resolution = len(qtm_id)
        num_edges = 3
        qtm_feature = geodesic_dggs_to_feature(
            "qtm", qtm_id, resolution, cell_polygon, num_edges
        )
        return qtm_feature


def csv2qtm(csv_file, column="qtm"):
    """Convert CSV file containing QTM cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            qtm_id = row[column]
            qtm_feature = qtm2feature(qtm_id)
            if qtm_feature:
                geojson_features.append(qtm_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid QTM cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2qtm_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with qtm column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with qtm column")
    parser.add_argument(
        "-id", dest="id", help="Name of the qtm column (default: 'qtm')", default="qtm"
    )
    args = parser.parse_args()
    qtm_csv = args.csv
    qtm_id = args.id

    qtm_geojson = csv2qtm(qtm_csv, qtm_id)
    geojson_name = os.path.splitext(os.path.basename(qtm_csv))[0]
    geojson_path = f"{geojson_name}2qtm.geojson"

    with open(geojson_path, "w") as f:
        json.dump(qtm_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  OLC
#################################################################################
def olc2feature(olc_id):
    coord = olc.decode(olc_id)
    if coord:
        # Create the bounding box coordinates for the polygon
        min_lat, min_lon = coord.latitudeLo, coord.longitudeLo
        max_lat, max_lon = coord.latitudeHi, coord.longitudeHi
        resolution = coord.codeLength

        # Define the polygon based on the bounding box
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )
        olc_feature = graticule_dggs_to_feature("olc", olc_id, resolution, cell_polygon)
        return olc_feature


def csv2olc(csv_file, column="olc"):
    """Convert CSV file containing OLC cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            olc_id = row[column]
            olc_feature = olc2feature(olc_id)
            if olc_feature:
                geojson_features.append(olc_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid OLC cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2olc_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with OLC column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with OLC column")
    parser.add_argument(
        "-id", dest="id", help="Name of the OLC column (default: 'olc')", default="olc"
    )
    args = parser.parse_args()
    olc_csv = args.csv
    olc_id = args.id

    olc_geojson = csv2olc(olc_csv, olc_id)
    geojson_name = os.path.splitext(os.path.basename(olc_csv))[0]
    geojson_path = f"{geojson_name}2olc.geojson"

    with open(geojson_path, "w") as f:
        json.dump(olc_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Geohash
#################################################################################
def geohash2feature(geohash_id):
    bbox = geohash.bbox(geohash_id)
    if bbox:
        min_lat, min_lon = bbox["s"], bbox["w"]  # Southwest corner
        max_lat, max_lon = bbox["n"], bbox["e"]  # Northeast corner
        resolution = len(geohash_id)

        # Define the polygon based on the bounding box
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )
        geohash_feature = graticule_dggs_to_feature(
            "geohash", geohash_id, resolution, cell_polygon
        )
        return geohash_feature


def csv2geohash(csv_file, column="geohash"):
    """Convert CSV file containing Geohash cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            geohash_id = row[column]
            geohash_feature = geohash2feature(geohash_id)
            if geohash_feature:
                geojson_features.append(geohash_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid Geohash cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2geohash_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with Geohash column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with Geohash column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the Geohash column (default: 'geohash')",
        default="geohash",
    )
    args = parser.parse_args()
    geohash_csv = args.csv
    geohash_id = args.id

    geohash_geojson = csv2geohash(geohash_csv, geohash_id)
    geojson_name = os.path.splitext(os.path.basename(geohash_csv))[0]
    geojson_path = f"{geojson_name}2geohash.geojson"

    with open(geojson_path, "w") as f:
        json.dump(geohash_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  GEOREF
#################################################################################
def georef2feature(georef_id):
    # Need to check georef.georefcell(georef_id) function
    center_lat, center_lon, min_lat, min_lon, max_lat, max_lon, resolution = (
        georef.georefcell(georef_id)
    )
    if center_lat:
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )
        georef_feature = graticule_dggs_to_feature(
            "georef", georef_id, resolution, cell_polygon
        )
        return georef_feature


def csv2georef(csv_file, column="georef"):
    """Convert CSV file containing Georef cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            georef_id = row[column]
            georef_feature = georef2feature(georef_id)
            if georef_feature:
                geojson_features.append(georef_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid Georef cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2georef_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with GEOREF column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with GEOREF column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the GEOREF column (default: 'georef')",
        default="georef",
    )
    args = parser.parse_args()
    georef_csv = args.csv
    georef_id = args.id

    georef_geojson = csv2georef(georef_csv, georef_id)
    geojson_name = os.path.splitext(os.path.basename(georef_csv))[0]
    geojson_path = f"{geojson_name}2georef.geojson"

    with open(geojson_path, "w") as f:
        json.dump(georef_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  MGRS
#################################################################################
def mgrs2feature(mgrs_id):
    # Need to check if MGRS cell is intersectd by GZD
    min_lat, min_lon, max_lat, max_lon, resolution = mgrs.mgrscell(mgrs_id)
    if min_lat:
        # Define the polygon based on the bounding box
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )

        mgrs_feature = graticule_dggs_to_feature(
            "georef", mgrs_id, resolution, cell_polygon
        )
        return mgrs_feature


def csv2mgrs(csv_file, column="mgrs"):
    """Convert CSV file containing MGRS cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            mgrs_id = row[column]
            mgrs_feature = mgrs2feature(mgrs_id)
            if mgrs_feature:
                geojson_features.append(mgrs_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid MGRS cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2mgrs_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with mgrs column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with mgrs column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the mgrs column (default: 'mgrs')",
        default="mgrs",
    )
    args = parser.parse_args()
    mgrs_csv = args.csv
    mgrs_id = args.id

    mgrs_geojson = csv2mgrs(mgrs_csv, mgrs_id)
    geojson_name = os.path.splitext(os.path.basename(mgrs_csv))[0]
    geojson_path = f"{geojson_name}2mgrs.geojson"

    with open(geojson_path, "w") as f:
        json.dump(mgrs_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Tilecode
#################################################################################
def tilecode2feature(tilecode_id):
    # Extract z, x, y from the tilecode using regex
    match = re.match(r"z(\d+)x(\d+)y(\d+)", tilecode_id)
    # Convert matched groups to integers
    z = int(match.group(1))
    x = int(match.group(2))
    y = int(match.group(3))

    # Get the bounds of the tile in (west, south, east, north)
    bounds = mercantile.bounds(x, y, z)
    if bounds:
        # Create the bounding box coordinates for the polygon
        min_lat, min_lon = bounds.south, bounds.west
        max_lat, max_lon = bounds.north, bounds.east
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )
        if cell_polygon:
            resolution = z
            tilecode_feature = graticule_dggs_to_feature(
                "tilecode", tilecode_id, resolution, cell_polygon
            )
            return tilecode_feature


def csv2tilecode(csv_file, column="tilecode"):
    """Convert CSV file containing Tilecode cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            tilecode_id = row[column]
            tilecode_feature = tilecode2feature(tilecode_id)
            if tilecode_feature:
                geojson_features.append(tilecode_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid Tilecode cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2tilecode_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with tilecode column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with tilecode column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the tilecode column (default: 'tilecode')",
        default="tilecode",
    )
    args = parser.parse_args()
    tilecode_csv = args.csv
    tilecode_id = args.id

    tilecode_geojson = csv2tilecode(tilecode_csv, tilecode_id)
    geojson_name = os.path.splitext(os.path.basename(tilecode_csv))[0]
    geojson_path = f"{geojson_name}2tilecode.geojson"

    with open(geojson_path, "w") as f:
        json.dump(tilecode_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Quadkey
#################################################################################
def quadkey2feature(quadkey_id):
    tile = mercantile.quadkey_to_tile(quadkey_id)
    # Format as tilecode
    z = tile.z
    x = tile.x
    y = tile.y
    # Get the bounds of the tile in (west, south, east, north)
    bounds = mercantile.bounds(x, y, z)

    if bounds:
        # Create the bounding box coordinates for the polygon
        min_lat, min_lon = bounds.south, bounds.west
        max_lat, max_lon = bounds.north, bounds.east

        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )

        resolution = z
        quadkey_feature = graticule_dggs_to_feature(
            "quadkey", quadkey_id, resolution, cell_polygon
        )
        return quadkey_feature


def csv2quadkey(csv_file, column="quadkey"):
    """Convert CSV file containing Quadkey cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            quadkey_id = row[column]
            quadkey_feature = quadkey2feature(quadkey_id)
            if quadkey_feature:
                geojson_features.append(quadkey_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid Quadkey cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2quadkey_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with quadkey column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with quadkey column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the quadkey column (default: 'quadkey')",
        default="quadkey",
    )
    args = parser.parse_args()
    quadkey_csv = args.csv
    quadkey_id = args.id

    quadkey_geojson = csv2quadkey(quadkey_csv, quadkey_id)
    geojson_name = os.path.splitext(os.path.basename(quadkey_csv))[0]
    geojson_path = f"{geojson_name}2quadkey.geojson"

    with open(geojson_path, "w") as f:
        json.dump(quadkey_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  Maidenhead
#################################################################################
def maidenhead2feature(maidenhead_id):
    # Decode the Open Location Code into a CodeArea object
    _, _, min_lat, min_lon, max_lat, max_lon, _ = maidenhead.maidenGrid(maidenhead_id)
    if min_lat:
        resolution = int(len(maidenhead_id) / 2)
        # Define the polygon based on the bounding box
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )
        maidenhead_feature = graticule_dggs_to_feature(
            "gars", maidenhead_id, resolution, cell_polygon
        )
        return maidenhead_feature


def csv2maidenhead(csv_file, column="maidenhead"):
    """Convert CSV file containing Maidenhead cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            maidenhead_id = row[column]
            maidenhead_feature = maidenhead2feature(maidenhead_id)
            if maidenhead_feature:
                geojson_features.append(maidenhead_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid Maidenhead cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2maidenhead_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with Maidenhead column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with Maidenhead column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the Maidenhead column (default: 'maidenhead')",
        default="maidenhead",
    )
    args = parser.parse_args()
    maidenhead_csv = args.csv
    maidenhead_id = args.id

    maidenhead_geojson = csv2maidenhead(maidenhead_csv, maidenhead_id)
    geojson_name = os.path.splitext(os.path.basename(maidenhead_csv))[0]
    geojson_path = f"{geojson_name}2maidenhead.geojson"

    with open(geojson_path, "w") as f:
        json.dump(maidenhead_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")


#################################################################################
#  GARS
#################################################################################
def gars2feature(gars_id):
    gars_grid = garsgrid.GARSGrid(gars_id)
    wkt_polygon = gars_grid.polygon

    if wkt_polygon:
        # Convert minute-based resolution to 1-4 scale
        resolution_minute = gars_grid.resolution
        if resolution_minute == 30:
            resolution = 1
        elif resolution_minute == 15:
            resolution = 2
        elif resolution_minute == 5:
            resolution = 3
        elif resolution_minute == 1:
            resolution = 4
        else:
            resolution = 1  # Default to level 1 if unknown

        cell_polygon = Polygon(list(wkt_polygon.exterior.coords))
        gars_feature = graticule_dggs_to_feature(
            "gars", gars_id, resolution, cell_polygon
        )
        return gars_feature


def csv2gars(csv_file, column="gars"):
    """Convert CSV file containing GARS cell IDs to GeoJSON."""
    try:
        df = pd.read_csv(csv_file)
    except Exception as e:
        parsed = urlparse(csv_file)
        if not parsed.scheme and not os.path.exists(csv_file):
            print(f"Error: Input file {csv_file} does not exist.")
            return None
        print(f"Error reading CSV file: {e}")
        return None

    if column not in df.columns:
        print(
            f"Error: Column '{column}' is missing in the input CSV. Please check and try again."
        )
        return None

    geojson_features = []
    for _, row in df.iterrows():
        try:
            gars_id = row[column]
            gars_feature = gars2feature(gars_id)
            if gars_feature:
                geojson_features.append(gars_feature)
        except Exception as e:
            print(f"Error processing row: {str(e)}")
            continue

    if not geojson_features:
        print("No valid GARS cells found in the CSV file.")
        return None

    geojson = {"type": "FeatureCollection", "features": geojson_features}
    return geojson


def csv2gars_cli():
    parser = argparse.ArgumentParser(
        description="Convert CSV with GARS column to GeoJSON"
    )
    parser.add_argument("csv", help="Input CSV file with GARS column")
    parser.add_argument(
        "-id",
        dest="id",
        help="Name of the GARS column (default: 'gars')",
        default="gars",
    )
    args = parser.parse_args()
    gars_csv = args.csv
    gars_id = args.id

    gars_geojson = csv2gars(gars_csv, gars_id)
    geojson_name = os.path.splitext(os.path.basename(gars_csv))[0]
    geojson_path = f"{geojson_name}2gars.geojson"

    with open(geojson_path, "w") as f:
        json.dump(gars_geojson, f)

    print(f"GeoJSON saved to {geojson_path}")



from vgrid.utils import s2, olc, geohash, georef, mgrs, mercantile, maidenhead
from vgrid.utils.gars import garsgrid
from vgrid.utils.qtm import constructGeometry, qtm_id_to_facet
import h3

from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS
from vgrid.utils.rhealpixdggs.utils import my_round
from vgrid.utils.rhealpixdggs.ellipsoids import WGS84_ELLIPSOID
import platform

if platform.system() == "Windows":
    from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
    from vgrid.utils.eaggr.enums.model import Model
    from vgrid.generator.isea4tgrid import fix_isea4t_wkt, fix_isea4t_antimeridian_cells
    from vgrid.generator.isea3hgrid import isea3h_cell_to_polygon

if platform.system() == "Linux":
    from vgrid.utils.dggrid4py import DGGRIDv7, dggs_types


from vgrid.utils.easedggs.constants import levels_specs
from vgrid.utils.easedggs.dggs.grid_addressing import grid_ids_to_geos

from shapely.wkt import loads
from shapely.geometry import shape, Polygon, mapping

import json
import re
import os
import argparse
from vgrid.generator.h3grid import fix_h3_antimeridian_cells
from vgrid.generator.rhealpixgrid import fix_rhealpix_antimeridian_cells

from vgrid.utils.antimeridian import fix_polygon

from vgrid.generator.settings import (
    graticule_dggs_to_feature,
    geodesic_dggs_to_feature,
    isea3h_accuracy_res_dict,
)

from pyproj import Geod

geod = Geod(ellps="WGS84")
E = WGS84_ELLIPSOID


def h32geojson(h3_ids):
    """
    Convert a list of H3 cell IDs to a GeoJSON FeatureCollection.
    Accepts a single h3_id (string) or a list of h3_ids.
    Skips invalid or error-prone cells.
    """
    if isinstance(h3_ids, str):
        h3_ids = [h3_ids]
    h3_features = []
    for h3_id in h3_ids:
        try:
            cell_boundary = h3.cell_to_boundary(h3_id)
            if cell_boundary:
                filtered_boundary = fix_h3_antimeridian_cells(cell_boundary)
                reversed_boundary = [(lon, lat) for lat, lon in filtered_boundary]
                cell_polygon = Polygon(reversed_boundary)
                resolution = h3.get_resolution(h3_id)
                num_edges = 6
                if h3.is_pentagon(h3_id):
                    num_edges = 5
                h3_feature = geodesic_dggs_to_feature(
                    "h3", h3_id, resolution, cell_polygon, num_edges
                )
                h3_features.append(h3_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": h3_features}


def h32geojson_cli():
    """
    Command-line interface for h32geojson supporting multiple H3 cell IDs.
    """
    parser = argparse.ArgumentParser(description="Convert H3 cell ID(s) to GeoJSON")
    parser.add_argument(
        "h3",
        nargs="+",
        help="Input H3 cell ID(s), e.g., h32geojson 8d65b56628e46bf 8d65b56628e46c3",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(h32geojson(args.h3))
    print(geojson_data)


def s22geojson(s2_tokens):
    """
    Convert a list of S2 cell tokens to a GeoJSON FeatureCollection.
    Accepts a single s2_token (string) or a list of s2_tokens.
    Skips invalid or error-prone cells.
    """
    if isinstance(s2_tokens, str):
        s2_tokens = [s2_tokens]
    s2_features = []
    for s2_token in s2_tokens:
        try:
            cell_id = s2.CellId.from_token(s2_token)
            cell = s2.Cell(cell_id)
            if cell:
                vertices = [cell.get_vertex(i) for i in range(4)]
                shapely_vertices = []
                for vertex in vertices:
                    lat_lng = s2.LatLng.from_point(vertex)
                    longitude = lat_lng.lng().degrees
                    latitude = lat_lng.lat().degrees
                    shapely_vertices.append((longitude, latitude))
                shapely_vertices.append(shapely_vertices[0])
                cell_polygon = fix_polygon(Polygon(shapely_vertices))
                resolution = cell_id.level()
                num_edges = 4
                s2_feature = geodesic_dggs_to_feature(
                    "s2", s2_token, resolution, cell_polygon, num_edges
                )
                s2_features.append(s2_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": s2_features}


def s22geojson_cli():
    """
    Command-line interface for s22geojson supporting multiple S2 cell tokens.
    """
    parser = argparse.ArgumentParser(description="Convert S2 cell token(s) to GeoJSON")
    parser.add_argument(
        "s2",
        nargs="+",
        help="Input S2 cell token(s), e.g., s22geojson 31752f45cc94 31752f45cc95",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(s22geojson(args.s2))
    print(geojson_data)


def rhealpix_cell_to_polygon(cell):
    vertices = [
        tuple(my_round(coord, 14) for coord in vertex)
        for vertex in cell.vertices(plane=False)
    ]
    if vertices[0] != vertices[-1]:
        vertices.append(vertices[0])
    vertices = fix_rhealpix_antimeridian_cells(vertices)
    return Polygon(vertices)


def rhealpix2geojson(rhealpix_ids):
    if isinstance(rhealpix_ids, str):
        rhealpix_ids = [rhealpix_ids]
    rhealpix_features = []
    for rhealpix_id in rhealpix_ids:
        try:
            rhealpix_uids = (rhealpix_id[0],) + tuple(map(int, rhealpix_id[1:]))
            rhealpix_dggs = RHEALPixDGGS(
                ellipsoid=E, north_square=1, south_square=3, N_side=3
            )
            rhealpix_cell = rhealpix_dggs.cell(rhealpix_uids)
            if rhealpix_cell:
                resolution = rhealpix_cell.resolution
                cell_polygon = rhealpix_cell_to_polygon(rhealpix_cell)
                num_edges = 4
                if rhealpix_cell.ellipsoidal_shape() == "dart":
                    num_edges = 3
                rhealpix_feature = geodesic_dggs_to_feature(
                    "rhealpix", rhealpix_id, resolution, cell_polygon, num_edges
                )
                rhealpix_features.append(rhealpix_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": rhealpix_features}


def rhealpix2geojson_cli():
    """
    Command-line interface for rhealpix2geojson supporting multiple Rhealpix cell IDs.
    """
    parser = argparse.ArgumentParser(
        description="Convert Rhealpix cell ID(s) to GeoJSON"
    )
    parser.add_argument(
        "rhealpix",
        nargs="+",
        help="Input Rhealpix cell ID(s), e.g., rhealpix2geojson R31260335553825 R31260335553826",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(rhealpix2geojson(args.rhealpix))
    print(geojson_data)


def isea4t2geojson(isea4t_ids):
    if isinstance(isea4t_ids, str):
        isea4t_ids = [isea4t_ids]
    isea4t_features = []
    if platform.system() == "Windows":
        for isea4t_id in isea4t_ids:
            try:
                isea4t_dggs = Eaggr(Model.ISEA4T)
                cell_to_shape = isea4t_dggs.convert_dggs_cell_outline_to_shape_string(
                    DggsCell(isea4t_id), ShapeStringFormat.WKT
                )
                cell_to_shape_fixed = loads(fix_isea4t_wkt(cell_to_shape))
                if (
                    isea4t_id.startswith("00")
                    or isea4t_id.startswith("09")
                    or isea4t_id.startswith("14")
                    or isea4t_id.startswith("04")
                    or isea4t_id.startswith("19")
                ):
                    cell_to_shape_fixed = fix_isea4t_antimeridian_cells(
                        cell_to_shape_fixed
                    )
                if cell_to_shape_fixed:
                    resolution = len(isea4t_id) - 2
                    num_edges = 3
                    cell_polygon = Polygon(list(cell_to_shape_fixed.exterior.coords))
                    isea4t_feature = geodesic_dggs_to_feature(
                        "isea4t", isea4t_id, resolution, cell_polygon, num_edges
                    )
                    isea4t_features.append(isea4t_feature)
            except Exception:
                continue
        return {"type": "FeatureCollection", "features": isea4t_features}


def isea4t2geojson_cli():
    """
    Command-line interface for isea4t2geojson supporting multiple ISEA4T cell IDs.
    """
    parser = argparse.ArgumentParser(
        description="Convert Open-Eaggr ISEA4T cell ID(s) to GeoJSON"
    )
    parser.add_argument(
        "isea4t",
        nargs="+",
        help="Input isea4t code(s), e.g., isea4t2geojson 131023133313201333311333 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(isea4t2geojson(args.isea4t))
    print(geojson_data)


def isea3h2geojson(isea3h_ids):
    if isinstance(isea3h_ids, str):
        isea3h_ids = [isea3h_ids]
    features = []
    if platform.system() == "Windows":
        for isea3h_id in isea3h_ids:
            try:
                isea3h_dggs = Eaggr(Model.ISEA3H)
                isea3h_cell = DggsCell(isea3h_id)
                cell_polygon = isea3h_cell_to_polygon(isea3h_dggs, isea3h_cell)
                cell_centroid = cell_polygon.centroid
                center_lat = round(cell_centroid.y, 7)
                center_lon = round(cell_centroid.x, 7)
                cell_area = round(abs(geod.geometry_area_perimeter(cell_polygon)[0]), 3)
                cell_perimeter = abs(geod.geometry_area_perimeter(cell_polygon)[1])
                isea3h2point = isea3h_dggs.convert_dggs_cell_to_point(isea3h_cell)
                cell_accuracy = isea3h2point._accuracy
                avg_edge_len = cell_perimeter / 6
                cell_resolution = isea3h_accuracy_res_dict.get(cell_accuracy)
                if cell_resolution == 0:
                    avg_edge_len = cell_perimeter / 3
                if cell_accuracy == 0.0:
                    if round(avg_edge_len, 2) == 0.06:
                        cell_resolution = 33
                    elif round(avg_edge_len, 2) == 0.03:
                        cell_resolution = 34
                    elif round(avg_edge_len, 2) == 0.02:
                        cell_resolution = 35
                    elif round(avg_edge_len, 2) == 0.01:
                        cell_resolution = 36
                    elif round(avg_edge_len, 3) == 0.007:
                        cell_resolution = 37
                    elif round(avg_edge_len, 3) == 0.004:
                        cell_resolution = 38
                    elif round(avg_edge_len, 3) == 0.002:
                        cell_resolution = 39
                    elif round(avg_edge_len, 3) <= 0.001:
                        cell_resolution = 40
                feature = {
                    "type": "Feature",
                    "geometry": mapping(cell_polygon),
                    "properties": {
                        "isea3h": isea3h_id,
                        "resolution": cell_resolution,
                        "center_lat": center_lat,
                        "center_lon": center_lon,
                        "avg_edge_len": round(avg_edge_len, 3),
                        "cell_area": cell_area,
                    },
                }
                features.append(feature)
            except Exception:
                continue
        feature_collection = {"type": "FeatureCollection", "features": features}
        return feature_collection


def isea3h2geojson_cli():
    """
    Command-line interface for isea3h2geojson supporting multiple ISEA3H cell IDs.
    """
    parser = argparse.ArgumentParser(description="Convert ISEA3H ID(s) to GeoJSON")
    parser.add_argument(
        "isea3h",
        nargs="+",
        help="Input ISEA3H cell ID(s), e.g., isea3h2geojson 1327916769,-55086 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(isea3h2geojson(args.isea3h))
    print(geojson_data)


def dggrid2geojson(dggrid_id, dggs_type, resolution):
    if platform.system() == "Linux":
        dggrid_instance = DGGRIDv7(
            executable="/usr/local/bin/dggrid",
            working_dir=".",
            capture_logs=False,
            silent=True,
            tmp_geo_out_legacy=False,
            debug=False,
        )
        dggrid_cell = dggrid_instance.grid_cell_polygons_from_cellids(
            [dggrid_id], dggs_type, resolution, split_dateline=True
        )

        gdf = dggrid_cell.set_geometry("geometry")  # Ensure the geometry column is set
        # Check and set CRS to EPSG:4326 if needed
        if gdf.crs is None:
            gdf.set_crs(epsg=4326, inplace=True)
        elif not gdf.crs.equals("EPSG:4326"):
            gdf = gdf.to_crs(epsg=4326)

        feature_collection = gdf.to_json()
        return feature_collection


def dggrid2geojson_cli():
    """
    Command-line interface for dggrid2geojson.
    """
    parser = argparse.ArgumentParser(
        description="Convert DGGRID code to GeoJSON. \
                                     Usage: dggrid2geojson <SEQNUM> <dggs_type> <res>. \
                                     Ex: dggrid2geojson 783229476878 ISEA7H 13"
    )
    parser.add_argument("dggrid", help="Input DGGRID code in SEQNUM format")
    parser.add_argument(
        "type",
        choices=dggs_types,
        help="Select a DGGS type from the available options.",
    )
    parser.add_argument("res", type=int, help="resolution")
    # parser.add_argument("address", choices=input_address_types, help="Address type")

    args = parser.parse_args()
    geojson_data = dggrid2geojson(args.dggrid, args.type, args.res)
    print(geojson_data)


def ease2geojson(ease_ids):
    if isinstance(ease_ids, str):
        ease_ids = [ease_ids]
    ease_features = []
    for ease_id in ease_ids:
        try:
            level = int(ease_id[1])
            level_spec = levels_specs[level]
            n_row = level_spec["n_row"]
            n_col = level_spec["n_col"]
            geo = grid_ids_to_geos([ease_id])
            center_lon, center_lat = geo["result"]["data"][0]
            cell_min_lat = center_lat - (180 / (2 * n_row))
            cell_max_lat = center_lat + (180 / (2 * n_row))
            cell_min_lon = center_lon - (360 / (2 * n_col))
            cell_max_lon = center_lon + (360 / (2 * n_col))
            cell_polygon = Polygon(
                [
                    [cell_min_lon, cell_min_lat],
                    [cell_max_lon, cell_min_lat],
                    [cell_max_lon, cell_max_lat],
                    [cell_min_lon, cell_max_lat],
                    [cell_min_lon, cell_min_lat],
                ]
            )
            resolution = level
            num_edges = 4
            ease_feature = geodesic_dggs_to_feature(
                "ease", ease_id, resolution, cell_polygon, num_edges
            )
            ease_features.append(ease_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": ease_features}


def ease2geojson_cli():
    """
    Command-line interface for ease2geojson supporting multiple EASE-DGGS codes.
    """
    parser = argparse.ArgumentParser(description="Convert EASE-DGGS code(s) to GeoJSON")
    parser.add_argument(
        "ease",
        nargs="+",
        help="Input EASE-DGGS code(s), e.g., ease2geojson L4.165767.02.02.20.71 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(ease2geojson(args.ease))
    print(geojson_data)


def qtm2geojson(qtm_ids):
    """
    Convert a list of QTM cell IDs to a GeoJSON FeatureCollection.
    Accepts a single qtm_id (string) or a list of qtm_ids.
    Skips invalid or error-prone cells.
    """
    if isinstance(qtm_ids, str):
        qtm_ids = [qtm_ids]
    qtm_features = []
    for qtm_id in qtm_ids:
        try:
            facet = qtm_id_to_facet(qtm_id)
            cell_polygon = constructGeometry(facet)
            resolution = len(qtm_id)
            num_edges = 3
            qtm_feature = geodesic_dggs_to_feature(
                "qtm", qtm_id, resolution, cell_polygon, num_edges
            )
            qtm_features.append(qtm_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": qtm_features}


def qtm2geojson_cli():
    """
    Command-line interface for qtm2geojson supporting multiple QTM cell IDs.
    """
    parser = argparse.ArgumentParser(description="Convert QTM cell ID(s) to GeoJSON")
    parser.add_argument(
        "qtm",
        nargs="+",
        help="Input QTM cell ID(s), e.g., qtm2geojson 42012321 42012322",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(qtm2geojson(args.qtm))
    print(geojson_data)


def olc2geojson(olc_id):
    # Decode the Open Location Code into a CodeArea object
    coord = olc.decode(olc_id)
    if coord:
        # Create the bounding box coordinates for the polygon
        min_lat, min_lon = coord.latitudeLo, coord.longitudeLo
        max_lat, max_lon = coord.latitudeHi, coord.longitudeHi
        resolution = coord.codeLength

        # Define the polygon based on the bounding box
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )
        olc_features = []
        olc_feature = graticule_dggs_to_feature("olc", olc_id, resolution, cell_polygon)
        olc_features.append(olc_feature)

    return {"type": "FeatureCollection", "features": olc_features}


def olc2geojson_cli():
    """
    Command-line interface for olc2geojson supporting multiple OLC codes.
    """
    parser = argparse.ArgumentParser(
        description="Convert OLC/ Google Plus Codes to GeoJSON"
    )
    parser.add_argument(
        "olc",
        nargs="+",
        help="Input OLC(s), e.g., olc2geojson 7P28QPG4+4P7 7P28QPG4+4P8",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(olc2geojson(args.olc))
    print(geojson_data)


def geohash2geojson(geohash_ids):
    if isinstance(geohash_ids, str):
        geohash_ids = [geohash_ids]
    geohash_features = []
    for geohash_id in geohash_ids:
        try:
            bbox = geohash.bbox(geohash_id)
            if bbox:
                min_lat, min_lon = bbox["s"], bbox["w"]
                max_lat, max_lon = bbox["n"], bbox["e"]
                resolution = len(geohash_id)
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                geohash_feature = graticule_dggs_to_feature(
                    "geohash", geohash_id, resolution, cell_polygon
                )
                geohash_features.append(geohash_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": geohash_features}


def geohash2geojson_cli():
    """
    Command-line interface for geohash2geojson supporting multiple Geohash cell IDs.
    """
    parser = argparse.ArgumentParser(
        description="Convert Geohash cell ID(s) to GeoJSON"
    )
    parser.add_argument(
        "geohash",
        nargs="+",
        help="Input Geohash cell ID(s), e.g., geohash2geojson w3gvk1td8 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(geohash2geojson(args.geohash))
    print(geojson_data)


def mgrs2geojson(mgrs_ids):
    if isinstance(mgrs_ids, str):
        mgrs_ids = [mgrs_ids]
    mgrs_features = []
    for mgrs_id in mgrs_ids:
        try:
            min_lat, min_lon, max_lat, max_lon, resolution = mgrs.mgrscell(mgrs_id)
            cell_polygon = Polygon(
                [
                    (min_lon, min_lat),
                    (max_lon, min_lat),
                    (max_lon, max_lat),
                    (min_lon, max_lat),
                    (min_lon, min_lat),
                ]
            )
            mgrs_feature = graticule_dggs_to_feature(
                "mgrs", mgrs_id, resolution, cell_polygon
            )
            try:
                gzd_json_path = os.path.join(
                    os.path.dirname(__file__), "../generator/gzd.geojson"
                )
                with open(gzd_json_path, "r") as f:
                    gzd_data = json.load(f)
                gzd_features = gzd_data["features"]
                gzd_feature = [
                    feature
                    for feature in gzd_features
                    if feature["properties"].get("gzd") == mgrs_id[:3]
                ][0]
                gzd_geom = shape(gzd_feature["geometry"])
                if mgrs_id[2] not in {"A", "B", "Y", "Z"}:
                    if cell_polygon.intersects(gzd_geom) and not gzd_geom.contains(
                        cell_polygon
                    ):
                        intersected_polygon = cell_polygon.intersection(gzd_geom)
                        if intersected_polygon:
                            mgrs_feature = graticule_dggs_to_feature(
                                "mgrs", mgrs_id, resolution, intersected_polygon
                            )
            except Exception:
                pass
            mgrs_features.append(mgrs_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": mgrs_features}


def mgrs2geojson_cli():
    """
    Command-line interface for mgrs2geojson supporting multiple MGRS cell IDs.
    """
    parser = argparse.ArgumentParser(description="Convert MGRS cell ID(s) to GeoJSON")
    parser.add_argument(
        "mgrs",
        nargs="+",
        help="Input MGRS cell ID(s), e.g., mgrs2geojson 48PXS866916 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(mgrs2geojson(args.mgrs))
    print(geojson_data)


def georef2geojson(georef_ids):
    if isinstance(georef_ids, str):
        georef_ids = [georef_ids]
    georef_features = []
    for georef_id in georef_ids:
        try:
            center_lat, center_lon, min_lat, min_lon, max_lat, max_lon, resolution = (
                georef.georefcell(georef_id)
            )
            if center_lat:
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                georef_feature = graticule_dggs_to_feature(
                    "georef", georef_id, resolution, cell_polygon
                )
                georef_features.append(georef_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": georef_features}


def georef2geojson_cli():
    """
    Command-line interface for georef2geojson supporting multiple GEOREF codes.
    """
    parser = argparse.ArgumentParser(description="Convert GEOREF code(s) to GeoJSON")
    parser.add_argument(
        "georef",
        nargs="+",
        help="Input GEOREF code(s), e.g., georef2geojson VGBL42404651 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(georef2geojson(args.georef))
    print(geojson_data)


def tilecode2geojson(tilecode_ids):
    if isinstance(tilecode_ids, str):
        tilecode_ids = [tilecode_ids]
    tilecode_features = []
    for tilecode_id in tilecode_ids:
        try:
            match = re.match(r"z(\d+)x(\d+)y(\d+)", tilecode_id)
            if not match:
                continue
            z = int(match.group(1))
            x = int(match.group(2))
            y = int(match.group(3))
            bounds = mercantile.bounds(x, y, z)
            if bounds:
                min_lat, min_lon = bounds.south, bounds.west
                max_lat, max_lon = bounds.north, bounds.east
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                resolution = z
                tilecode_feature = graticule_dggs_to_feature(
                    "tilecode_id", tilecode_id, resolution, cell_polygon
                )
                tilecode_features.append(tilecode_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": tilecode_features}


def tilecode2geojson_cli():
    """
    Command-line interface for tilecode2geojson supporting multiple Tilecodes.
    """
    parser = argparse.ArgumentParser(description="Convert Tilecode(s) to GeoJSON")
    parser.add_argument(
        "tilecode_id", nargs="+", help="Input Tilecode(s), e.g. z0x0y0 z1x1y1"
    )
    args = parser.parse_args()
    geojson_data = json.dumps(tilecode2geojson(args.tilecode_id))
    print(geojson_data)


def quadkey2geojson(quadkey_ids):
    if isinstance(quadkey_ids, str):
        quadkey_ids = [quadkey_ids]
    quadkey_features = []
    for quadkey_id in quadkey_ids:
        try:
            tile = mercantile.quadkey_to_tile(quadkey_id)
            z = tile.z
            x = tile.x
            y = tile.y
            bounds = mercantile.bounds(x, y, z)
            if bounds:
                min_lat, min_lon = bounds.south, bounds.west
                max_lat, max_lon = bounds.north, bounds.east
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                resolution = z
                quadkey_feature = graticule_dggs_to_feature(
                    "quadkey", quadkey_id, resolution, cell_polygon
                )
                quadkey_features.append(quadkey_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": quadkey_features}


def quadkey2geojson_cli():
    """
    Command-line interface for quadkey2geojson supporting multiple Quadkeys.
    """
    parser = argparse.ArgumentParser(description="Convert Quadkey(s) to GeoJSON")
    parser.add_argument(
        "quadkey", nargs="+", help="Input Quadkey(s), e.g. 13223011131020220011133 ..."
    )
    args = parser.parse_args()
    geojson_data = json.dumps(quadkey2geojson(args.quadkey))
    print(geojson_data)


def maidenhead2geojson(maidenhead_ids):
    if isinstance(maidenhead_ids, str):
        maidenhead_ids = [maidenhead_ids]
    maidenhead_features = []
    for maidenhead_id in maidenhead_ids:
        try:
            _, _, min_lat, min_lon, max_lat, max_lon, _ = maidenhead.maidenGrid(
                maidenhead_id
            )
            if min_lat:
                resolution = int(len(maidenhead_id) / 2)
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                maidenhead_feature = graticule_dggs_to_feature(
                    "maidenhead", maidenhead_id, resolution, cell_polygon
                )
                maidenhead_features.append(maidenhead_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": maidenhead_features}


def maidenhead2geojson_cli():
    """
    Command-line interface for maidenhead2geojson supporting multiple Maidenhead cell IDs.
    """
    parser = argparse.ArgumentParser(
        description="Convert Maidenhead cell ID(s) to GeoJSON"
    )
    parser.add_argument(
        "maidenhead",
        nargs="+",
        help="Input Maidenhead cell ID(s), e.g., maidenhead2geojson OK30is46 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(maidenhead2geojson(args.maidenhead))
    print(geojson_data)


def gars2geojson(gars_ids):
    if isinstance(gars_ids, str):
        gars_ids = [gars_ids]
    gars_features = []
    for gars_id in gars_ids:
        try:
            gars_grid = garsgrid.GARSGrid(gars_id)
            wkt_polygon = gars_grid.polygon
            if wkt_polygon:
                resolution_minute = gars_grid.resolution
                resolution = 1
                if resolution_minute == 30:
                    resolution = 1
                elif resolution_minute == 15:
                    resolution = 2
                elif resolution_minute == 5:
                    resolution = 3
                elif resolution_minute == 1:
                    resolution = 4
                cell_polygon = Polygon(list(wkt_polygon.exterior.coords))
                gars_feature = graticule_dggs_to_feature(
                    "gars", gars_id, resolution, cell_polygon
                )
                gars_features.append(gars_feature)
        except Exception:
            continue
    return {"type": "FeatureCollection", "features": gars_features}


def gars2geojson_cli():
    """
    Command-line interface for gars2geojson supporting multiple GARS cell IDs.
    """
    parser = argparse.ArgumentParser(description="Convert GARS cell ID(s) to GeoJSON")
    parser.add_argument(
        "gars",
        nargs="+",
        help="Input GARS cell ID(s), e.g., gars2geojson 574JK1918 ...",
    )
    args = parser.parse_args()
    geojson_data = json.dumps(gars2geojson(args.gars))
    print(geojson_data)



import os
import argparse
import json
from tqdm import tqdm
import rasterio
import numpy as np
from shapely.geometry import Polygon
from vgrid.stats.geohashstats import geohash_metrics
from vgrid.generator.settings import graticule_dggs_to_feature
from vgrid.conversion.latlon2dggs import latlon2geohash
from math import cos, radians
from vgrid.utils import geohash
import csv


def get_nearest_geohash_resolution(raster_path):
    with rasterio.open(raster_path) as src:
        transform = src.transform
        crs = src.crs
        pixel_width = transform.a
        pixel_height = -transform.e
        cell_size = pixel_width * pixel_height

        if crs.is_geographic:
            # Latitude of the raster center
            center_latitude = (src.bounds.top + src.bounds.bottom) / 2
            # Convert degrees to meters
            meter_per_degree_lat = 111_320  # Roughly 1 degree latitude in meters
            meter_per_degree_lon = meter_per_degree_lat * cos(radians(center_latitude))

            pixel_width_m = pixel_width * meter_per_degree_lon
            pixel_height_m = pixel_height * meter_per_degree_lat
            cell_size = pixel_width_m * pixel_height_m

    nearest_resolution = None
    min_diff = float("inf")

    # Check resolutions from 1 to 10
    for res in range(1, 11):
        _, _, avg_area = geohash_metrics(res)
        diff = abs(avg_area - cell_size)
        # If the difference is smaller than the current minimum, update the nearest resolution
        if diff < min_diff:
            min_diff = diff
            nearest_resolution = res

    return nearest_resolution


def convert_numpy_types(obj):
    """Recursively convert NumPy types to native Python types"""
    if isinstance(obj, np.generic):
        return obj.item()  # Convert numpy types like np.uint8 to native Python int
    elif isinstance(obj, dict):
        return {k: convert_numpy_types(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_numpy_types(v) for v in obj]
    else:
        return obj


def raster2geohash(raster_path, resolution=None, format="geojson"):
    # Step 1: Determine the nearest geohash resolution if none is provided
    if resolution is None:
        resolution = get_nearest_geohash_resolution(raster_path)
        print(f"Nearest geohash resolution determined: {resolution}")

    # Validate resolution is in valid range
    if resolution < 1 or resolution > 10:
        raise ValueError("Resolution must be in range [1..10]")

    # Open the raster file to get metadata and data
    with rasterio.open(raster_path) as src:
        raster_data = src.read()  # Read all bands
        transform = src.transform
        width, height = src.width, src.height
        band_count = src.count  # Number of bands in the raster

    geohash_ids = set()

    for row in range(height):
        for col in range(width):
            lon, lat = transform * (col, row)
            geohash_id = latlon2geohash(lat, lon, resolution)
            geohash_ids.add(geohash_id)

    # Sample the raster values at the centroids of the geohash hexagons
    geohash_data = []

    for geohash_id in tqdm(geohash_ids, desc="Resampling", unit=" cells"):
        # Get the centroid of the geohash cell
        centroid_lat, centroid_lon = geohash.decode(geohash_id)
        # Sample the raster values at the centroid (lat, lon)
        col, row = ~transform * (centroid_lon, centroid_lat)

        if 0 <= col < width and 0 <= row < height:
            # Get the values for all bands at this centroid
            values = raster_data[:, int(row), int(col)]
            geohash_data.append(
                {
                    "geohash": geohash_id,
                    **{
                        f"band_{i + 1}": values[i] for i in range(band_count)
                    },  # Create separate columns for each band
                }
            )

    if format.lower() == "csv":
        import io

        output = io.StringIO()
        if geohash_data:
            writer = csv.DictWriter(output, fieldnames=geohash_data[0].keys())
            writer.writeheader()
            writer.writerows(geohash_data)
        return output.getvalue()

    # Create the GeoJSON-like structure
    geohash_features = []
    for data in tqdm(geohash_data, desc="Converting to GeoJSON", unit=" cells"):
        geohash_id = data["geohash"]
        geohash_bbox = geohash.bbox(geohash_id)
        if geohash_bbox:
            min_lat, min_lon = geohash_bbox["s"], geohash_bbox["w"]  # Southwest corner
            max_lat, max_lon = geohash_bbox["n"], geohash_bbox["e"]  # Northeast corner
            cell_resolution = len(geohash_id)
            # Define the polygon based on the bounding box
            cell_polygon = Polygon(
                [
                    [min_lon, min_lat],  # Bottom-left corner
                    [max_lon, min_lat],  # Bottom-right corner
                    [max_lon, max_lat],  # Top-right corner
                    [min_lon, max_lat],  # Top-left corner
                    [min_lon, min_lat],  # Closing the polygon (same as the first point)
                ]
            )
            geohash_feature = graticule_dggs_to_feature(
                "geohash", geohash_id, cell_resolution, cell_polygon
            )
            band_properties = {
                f"band_{i + 1}": data[f"band_{i + 1}"] for i in range(band_count)
            }
            geohash_feature["properties"].update(convert_numpy_types(band_properties))
            geohash_features.append(geohash_feature)

    return {
        "type": "FeatureCollection",
        "features": geohash_features,
    }


def raster2geohash_cli():
    parser = argparse.ArgumentParser(
        description="Convert Raster in Geographic CRS to Geohash DGGS"
    )
    parser.add_argument("-raster", type=str, required=True, help="Raster file path")

    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        required=False,
        default=None,
        help="Resolution [1..10]",
        min=1,
        max=10,
    )

    parser.add_argument(
        "-f",
        "--format",
        type=str,
        required=False,
        default="geojson",
        choices=["geojson", "csv"],
        help="Output format (geojson or csv)",
    )

    args = parser.parse_args()
    raster = args.raster
    resolution = args.resolution
    output_format = args.format

    if not os.path.exists(raster):
        print(f"Error: The file {raster} does not exist.")
        return

    result = raster2geohash(raster, resolution, output_format)
    base_name = os.path.splitext(os.path.basename(raster))[0]

    if output_format.lower() == "csv":
        output_path = f"{base_name}2geohash.csv"
        # Get all possible field names from the first row
        fieldnames = list(result[0].keys())
        with open(output_path, "w", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(result)
    else:
        output_path = f"{base_name}2geohash.geojson"
        with open(output_path, "w") as f:
            json.dump(result, f)

    print(f"Output saved as {output_path}")



"""
Intiialize src directory

© Regents of the University of Minnesota. All rights reserved.
This software is released under an Apache 2.0 license. Further details about the Apache 2.0 license are available in the license.txt file.
"""

import pyproj

# enabling PROJ Content Delivery Network to ensure consistent results between PowerShell and Anaconda/Command Prompt
pyproj.network.set_network_enabled(active=True)



"""
Funtions for geographic transformations on to, and off of the EASE-DGGS.

© Regents of the University of Minnesota. All rights reserved.
This software is released under an Apache 2.0 license. Further details about the Apache 2.0 license are available in the license.txt file.
"""

from geopandas import GeoSeries
from shapely.geometry import Point

from vgrid.utils.easedggs.constants import (
    ease_crs,
    geo_crs,
)

from vgrid.utils.easedggs.dggs.utils import shift_range_ease, shift_range_grid_xy


def ease_coord_to_grid_xy_coord(x_ease, y_ease):
    """
    Convert DGGS cell coordinates to EASE Grid v2 coordinates

    Parameters
    ----------
    x : float
       EASE grid v2 x coordinate.
    y : float
        EASE grid v2 y coordinate.

    Returns
    -------
    (x_grid, y_grid)
    x_grid : float
        The transformed EASE Grid x cooridinate, in grid space units
    y_grid :  float
        The transformed EASE Grid x cooridinate, in grid space units
    """
    if not isinstance(x_ease, float) and not isinstance(y_ease, float):
        return False

    return (shift_range_ease(x_ease, "x"), shift_range_ease(y_ease, "y"))


def grid_xy_coord_to_ease_coord(x_grid, y_grid):
    """
    Convert EASE Grid v2 coordinates to DGGS cell

    Parameters
    ----------
    x_grid : float
       EASE-DGGS x coordinate.
    y_grid : float
        EASE-DGGS y coordinate.

    Returns
    -------
    (x_ease, y_ease) : tuple
        The transformed EASE-DGGS x,y cooridinate in as EASE v2 x,y coordinates

    """
    if not isinstance(x_grid, float) and not isinstance(y_grid, float):
        return False

    return (shift_range_grid_xy(x_grid, "x"), shift_range_grid_xy(y_grid, "y"))


def coords_ease_to_coords_grid(coords_ease, level=0):
    """
    Convert GeoSeries of EASE Grid v2 coordinates to EASE-DGGS coordinates

    Parameters
    ----------
    coords_ease : Geopandas GeoSeries
       GeoSeries of EASE Grid coordinate pair points (x y).
    level : str
        EASE-DGGS level.

    Returns
    -------
    coords_grid: Geopandas GeoSeries
        GeoSeries of transformed coodinante pair points in the EASE-DGGS coordinate system.
    """
    if not isinstance(coords_ease, GeoSeries):
        return False

    coords_grid = coords_ease.apply(
        lambda coord: Point(
            shift_range_ease(coord.x, "x"), shift_range_ease(coord.y, "y")
        )
    )

    return coords_grid


def coords_lon_lat_to_coords_ease(
    coords_lon_lat, source_crs=geo_crs, target_crs=ease_crs
):
    """
    Convience function for converting between geographic and EASE grid cooridnates.
    Parameters
    ----------
    coords_lon_lat : list
       List of coordinate pairs (longitude, latitude) of geographic cooridnates.
    source_crs : int
        The EPSG code for the source geographic coordinate pair. Default is 4326.
    target_crs : int
        The EPSG code for the target (EASE v2) coordinate pair. Default is ease_crs.

    Returns
    -------
    coords_ease : list
        List of transformed coordinate pairs (lon, lat), now in EASE Grid v2 cooridnates(ease_x, ease_y).
    """

    if not isinstance(coords_lon_lat, list):
        return False

    coords_lon_lat = GeoSeries(
        [Point(coord[0], coord[1]) for coord in coords_lon_lat], crs=source_crs
    )
    coords_ease = coords_lon_lat.to_crs(target_crs)

    return coords_ease



# Reference: https://github.com/pistell/Leaflet.DumbMGRS
# https://mgrs-mapper.com/app, https://military-history.fandom.com/wiki/Military_Grid_Reference_System
# https://codesandbox.io/s/how-to-toggle-react-leaflet-layer-control-and-rectangle-grid-f43xi?file=/src/App.js:1057-1309
# https://github.com/GeoSpark/gridoverlay/tree/d66ed86636c7ec3f02ca2e9298ac3086c2023f1d
# https://help.arcgis.com/en/arcgisdesktop/10.0/help/index.html#//00700000001n000000
# https://storymaps.arcgis.com/stories/842edf2b4381438b9a4edefed124775b
# https://github.com/dnlbaldwin/React-Leaflet-MGRS-Graticule
# https://dnlbaldwin.github.io/React-Leaflet-MGRS-Graticule/
# https://earth-info.nga.mil/index.php?dir=coordsys&action=mgrs-100km-polyline-dloads
# https://mgrs-data.org/metadata/
# https://ufl.maps.arcgis.com/apps/dashboards/2539764e24e74bd78f265f49c7adc2d1
# https://earth-info.nga.mil/index.php?dir=coordsys&action=gars-20x20-dloads

import json
from shapely.geometry import box, mapping
from tqdm import tqdm

bands = [
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
]


def generate_grid(polar=True):
    features = []

    def export_polygon(lon, lat, width, height, gzd):
        rect = box(lon, lat, lon + width, lat + height)
        features.append(
            {"type": "Feature", "geometry": mapping(rect), "properties": {"gzd": gzd}}
        )

    if polar:
        export_polygon(-180, -90, 180, 10, "A")
        export_polygon(0, -90, 180, 10, "B")

    lat = -80
    for b in tqdm(bands, desc="Generating Bands"):
        if b == "X":
            height = 12
            lon = -180
            for i in range(1, 31):
                gzd = "{:02d}{}".format(i, b)
                width = 6
                export_polygon(lon, lat, width, height, gzd)
                lon += width
            export_polygon(lon, lat, 9, height, "31X")
            lon += 9
            export_polygon(lon, lat, 12, height, "33X")
            lon += 12
            export_polygon(lon, lat, 12, height, "35X")
            lon += 12
            export_polygon(lon, lat, 9, height, "37X")
            lon += 9
            for i in range(38, 61):
                gzd = "{:02d}{}".format(i, b)
                width = 6
                export_polygon(lon, lat, width, height, gzd)
                lon += width
        else:
            height = 8
            lon = -180
            for i in range(1, 61):
                gzd = "{:02d}{}".format(i, b)
                if b == "V" and i == 31:
                    width = 3
                elif b == "V" and i == 32:
                    width = 9
                else:
                    width = 6
                export_polygon(lon, lat, width, height, gzd)
                lon += width
        lat += height

    if polar:
        export_polygon(-180, 84, 180, 6, "Y")
        export_polygon(0, 84, 180, 6, "Z")

    return features


def main():
    try:
        # Generate gzd grid
        features = generate_grid()

        # Save to GeoJSON
        geojson = {"type": "FeatureCollection", "features": features}
        with open("gzd.geojson", "w") as f:
            json.dump(geojson, f)

        print("GeoJSON saved to gzd.geojson")

    except Exception as e:
        print(f"An error occurred: {e}")


if __name__ == "__main__":
    main()



"""
GeoTransform functions relating to EASE-DGGS.

© Regents of the University of Minnesota. All rights reserved.
This software is released under an Apache 2.0 license. Further details about the Apache 2.0 license are available in the license.txt file.
"""
# © Regents of the University of Minnesota. All rights reserved.
# This software is released under an Apache 2.0 license. Further details about the Apache 2.0 license are available in the license.txt file.

import math

import numpy as np
import rasterio
from geopandas import GeoSeries
from itertools import chain, product
from pyproj import Transformer
from shapely.geometry import Point

from vgrid.utils.easedggs.constants import levels_specs, grid_spec, ease_crs, geo_crs

from vgrid.utils.easedggs.dggs.grid_addressing import (
    coords_ease_to_coords_grid,
    _grid_xy_to_grid_id,
)
from vgrid.utils.easedggs.dggs.utils import (
    shift_range_grid_multiple,
    pairwise_circle,
    flatten,
    add_nodes,
    epsilon_check,
)

######
#
# Helper functions
#
######
# define the pyproj Transforms necessary. need both a
#   from WGS84 -> EASE v2 (w2e) and
#   from EASE v2 -> WGS84 (e2w)
w2e = Transformer.from_crs(geo_crs, ease_crs, always_xy=True).transform
e2w = Transformer.from_crs(ease_crs, geo_crs, always_xy=True).transform


def grid_id_to_corner_coord(gid, level, shift=False):
    """Convert EASE-DGGS ID into an EASE v2 corner coordinate.

    Parameters
    ----------
    gid : str
        Grid ID of the cell.
    corner : str
        The type of corner to generate coordinates for
        (upper_left | upper_right | lower_right | lower_left)
    Returns
    ----------
    corner_x, corner_y : np.array
        EASE grid coordinate (x, y) of the cell' corner. Resolution determined using grid ID.
    """
    # r_digit = 6
    # decompse the EASE-DGGS ID into its level parts. Then remove the first compoenent
    #   which, is simply represent the Level of the cell.
    id_elements = gid.split(".")
    # level = int(id_elements[0][-1])
    id_elements.pop(0)

    # the gridID contains the index values for number of rows|columns, from a 0,0 orgin
    #   This is the 'global' position within the larger grid
    #
    #  Remember format is Lx.RRRCCC.RC.RC ...; y=R, x=C
    #
    #   Those R, C components indicate the location/number of grid cells that
    #   correspond with the location in ease grid space
    #
    #   This can be explicitly exploited such that:
    #   ([RRR, R, ... R] * [level[0]['x_length], ... level[6]['x_length]]).sum()
    #   ([CCC, C, ... C] * [level[0]['y_length],... level[6]['y_length]]).sum()
    #   will give the coordinates to the edge of the pixel. For example:
    #
    #   [L6.202482.13.21.00.00.00.00] would decompose into
    #     x = [482, 3, 2, 0, 0, 0, 0]
    #     y = [202, 1, 1, 0, 0, 0, 0]
    #
    #   To determining the coordinate of the left, top edges at Level 2
    #     x = 482 * level[0]['x_length] + 3 * [level[1]['x_length] + 2 * [level[`2`]['x_length]
    #     y = 202 * [level[0]['x_length] + 1 * [level[1]['x_length] + 1 * [level[2]['x_length]
    #
    #   and to determin if it is an edge at level 2, all the remaining Level components must be
    #       zero: [ 0, 0, 0, 0]] summing them all ensure that they are zero
    #
    #   if we asked for the Level 1 coordinate, it would be different. x|y coord is calculated
    #   using the L0, L1 compoents:
    #   x = 482 * level[0]['x_length] + 3 * [level[1]['x_length] + 2
    #
    #   but then we need to check if it is on an L1 edge. this requires all other compontes to be
    #       zeros:
    #       [2, 0, 0, 0, 0]
    #   since it is not a Level 1 edge, if we needed a right, bottom value, we'd have to shfit
    #       the last L1 index up by one

    # beyond level 0, the row | column index is grid location within the L0 pixes
    #    This is the relative position within the gloabl position.
    # spliting these into indexes, which requires converstion from str to ints
    level_x_i = np.array(
        [int(id[1]) if len(id) != 6 else int(id[3:]) for id in id_elements], dtype=int
    )
    level_y_i = np.array(
        [int(id[0]) if len(id) != 6 else int(id[0:3]) for id in id_elements], dtype=int
    )

    # remainders of levels are useful for figuring out if on edge on not
    #   depending on supplied level, we can drop unnecesary levels for
    #   determining the coodinates, but only need the levels < specified
    #   level for determing if it is an edge pixel or not
    if int(level) < 6:
        level_x_r = level_x_i[int(level) + 1 :]
        level_y_r = level_y_i[int(level) + 1 :]
    else:
        level_x_r = level_x_i[int(level) :]
        level_y_r = level_y_i[int(level) :]

    level_x_i = level_x_i[: int(level) + 1]
    level_y_i = level_y_i[: int(level) + 1]

    if (level_x_r.sum() != 0) and shift:
        # print(f'level_x_r: {level_x_r}; sum: {level_x_r.sum()}')
        # print('shifting x')
        level_x_i[-1] += 1

    if (level_y_r.sum() != 0) and shift:
        # print(f'level_y_r: {level_y_r}; sum: {level_y_r.sum()}')
        # print('shifting y')
        level_y_i[-1] += 1

    # print (f'level_x_i: {level_x_i}; level_y_i: {level_y_i}')
    corner_x = np.array(
        [level_x_i[i] * levels_specs[i]["x_length"] for i in range(level_x_i.shape[0])]
    ).sum()

    corner_y = np.array(
        [level_y_i[i] * levels_specs[i]["y_length"] for i in range(level_y_i.shape[0])]
    ).sum()

    return corner_x, corner_y


######
#
# main functions
#
######


def easedggs_grid_bounds(bounds, source_crs, level):
    """Determine EASE v2 coords corresponding to bounds of EASE-DGGS

    Parameters
    ----------
    bounds : tuple
        Tuple with the (min_x/left, min_y/bottom, max_x/right, max_y/top) of the soure raster.
    level : int
        EASE-DGGS Level of the
    source_crs : int
        EPSG code of coordinate reference systems for the source raster.

    Returns
    ----------
    easedggs_grid_bounds : tuple of floats
        Bounding box coordidinates with valid EASE-DGGS coordinates.
    """
    # first, convert the bounds coordinates into corner coordinates. these are tuples.
    #   next, convert the corners to 'line_segements'. these are the coordinate pairs
    #   connecting the corners
    #
    # Those line segmensts have 'duplicate' entries, since each line segment ends where next starts
    #   once we've delt with non-ease sourc_src below, we'll drop the final element of
    #   of line_segments, to reduce duplicates
    lower_left, upper_left, lower_right, upper_right = chain(
        product([bounds[0], bounds[2]], [bounds[1], bounds[3]])
    )
    # print(f'conrner passed in: {corners}')
    line_segments = list(
        pairwise_circle([upper_left, upper_right, lower_right, lower_left])
    )

    # if not in EASE v2, need to build a cood tranformer to get coords into EASE
    #   also need to add nodes to the each line segment. once that is done, we use
    #   the transformer to convert all individual nodes to EASE, while maintaining the
    #   original order.
    #   then use
    if source_crs != ease_crs:
        line_segments = [add_nodes(ls[0], ls[1]) for ls in line_segments]
        tranform_coords = Transformer.from_crs(
            source_crs, ease_crs, always_xy=True
        ).transform
        line_segments = [
            [tranform_coords(c[0], c[1]) for c in seg] for seg in line_segments
        ]

    # may not need to drop the last elements in each list (recall, the last element of
    #   one list, is the start of the next) since we'll just be using the nodes as Points
    #   in a geoseries. but if we wanted to make Polygon, the duplciate values pose
    #   problems
    line_segments = flatten([seg[:-1] for seg in line_segments])
    bound_box = GeoSeries([Point(c[0], c[1]) for c in line_segments], crs=ease_crs)
    bound_box = bound_box.total_bounds

    # now that we have ensured we have accounted for bounds not necessarily
    #   corresponding to vertices in EASE coordinates, we'll create the
    #   bounding box as vertices
    bound_box = GeoSeries(
        [
            Point(bound_box[0], bound_box[3]),  # upper_left
            Point(bound_box[2], bound_box[1]),
        ],  # lower_right
        crs=ease_crs,
    )

    bbox_grid_xy = coords_ease_to_coords_grid(bound_box)
    bbox_gridID = bbox_grid_xy.apply(lambda x: _grid_xy_to_grid_id(x, 6))

    # if we use upper_right, and lower left coords only, we can apply
    #   shift or not, to both points.
    upper_left = grid_id_to_corner_coord(bbox_gridID.iloc[0], level=level)
    lower_right = grid_id_to_corner_coord(bbox_gridID.iloc[1], level=level, shift=True)

    ease_xs = [upper_left[0], lower_right[0]]
    ease_ys = [lower_right[1], upper_left[1]]

    ease_xs = np.array(
        [shift_range_grid_multiple(ease_xs[i], axis="x") for i in range(len(ease_xs))]
    )
    ease_ys = np.array(
        [shift_range_grid_multiple(ease_ys[i], axis="y") for i in range(len(ease_ys))]
    )

    return [ease_xs[0], ease_ys[0], ease_xs[1], ease_ys[1]]  # , decimals=8)


def easedggs_align_check(in_raster, level):
    """Check if a raster is aligned to the EASE-DGGS.

    Takes a path to a raster and performs 6 tests of alignment with the
    EASE-DGGS. The first 4 tests ensure the grid is defined the same as
    the EASE-DGGS. If a raster is the same projection (EPSG:6933), is
    rectilinear (aligned to the axes), has square cells, and is one of
    the 7 EASE-DGGS resolutions, it is able to be aligned to the EASE-DGGS.
    The final 2 tests check that if you were to extrapolate your
    raster out to a global extent, that the upper left corner of the
    EASE-DGGS falls on the upper left corner of one of the input
    raster's extraploted cells.


    Parameters
    ------------
    in_raster: str, file object or pathlib.Path object
        Filename and path of input raster.

    level: int
        Integer from 0 to 6 corresponding to the EASE-DGGS levels
        (i.e., resolutions).

    Returns
    -------
    results: Dictionary
        Dictionary of test name and result of test (pass/fail).

    """
    if level not in [0, 1, 2, 3, 4, 5, 6]:
        raise Exception("Invalid grid level. Options are: 0, 1, 2, 3, 4, 5, 6")

    # Calculate the geotransform of the EASE-DGGS
    easedggs_t = rasterio.transform.from_origin(
        grid_spec["ease"]["min_x"],
        grid_spec["ease"]["max_y"],
        levels_specs[level]["x_length"],
        levels_specs[level]["x_length"],
    )

    with rasterio.open(in_raster, "r") as src:
        src_t = src.transform
        test_count = 6

        results = {
            "Projection is EASE-Grid 2.0 (EPSG:6933):": "Pass",
            "Dataset is rectilinear:": "Pass",
            "X and Y size are equal:": "Pass",
            "Cell size matches a EASE-DGGS resolution:": "Pass",
            "The X dimension of the cell corners matches a EASE-DGGS corner:": "Pass",
            "The Y dimension of the cell corners matches a EASE-DGGS corner:": "Pass",
            "Tests passed out of 6:": test_count,
        }

        if src.crs != "EPSG: 6933":
            results["Projection is EASE-Grid 2.0 (EPSG:6933):"] = str(
                "Fail. Supplied CRS is " + str(src.crs)
            )
            test_count -= 1

        if not src_t.is_rectilinear:
            results["Dataset is rectilinear:"] = "Fail"
            test_count -= 1

        if not epsilon_check(abs(src_t.a), abs(src_t.e)):
            results["X and Y size are equal:"] = "Fail"
            test_count -= 1

        if not epsilon_check(easedggs_t.a, src_t.a):
            results["Cell size matches a EASE-DGGS resolution:"] = "Fail"
            test_count -= 1

        # Find the input raster coordinates of the hypothetical input raster cell that overlaps the GEMS origin
        xr, yr = ~src_t * (easedggs_t.xoff, easedggs_t.yoff)

        # Move to the coordinates of the upper left corner of the above cell
        xrnd = math.floor(xr + 0.5) if xr > 0.0 else math.ceil(xr - 0.5)
        if epsilon_check(xrnd, xr):  # Test if on an edge, else go left
            xr = float(xrnd)
        else:
            xr = float(math.floor(xr))

        yrnd = math.floor(yr + 0.5) if yr > 0.0 else math.ceil(yr - 0.5)
        if epsilon_check(yrnd, yr):  # Test if on an edge, else go up
            yr = float(yrnd)
        else:
            yr = float(math.floor(yr))

        # Convert from input raster coordinates to world (ease) coordinates
        xw, yw = src_t * (xr, yr)

        # Compare x and y values calculated in previous step to gems origin coordinates
        if not epsilon_check(xw, easedggs_t.xoff):
            results[
                "The X dimension of the cell corners matches a EASE-DGGS corner:"
            ] = str("Fail. Difference = " + str(round(abs(xw - easedggs_t.xoff), 5)))
            test_count -= 1

        if not epsilon_check(yw, easedggs_t.yoff):
            results[
                "The Y dimension of the cell corners matches a EASE-DGGS corner:"
            ] = str("Fail. Difference = " + str(round(abs(yw - easedggs_t.yoff), 5)))
            test_count -= 1

        else:
            pass
        results["Tests passed out of 6:"] = str(test_count)

        # for item in results:
        #     print(item, results[item])

        return results



import sys
import argparse
import json
from tqdm import tqdm
from shapely.geometry import Polygon, box, MultiPoint, MultiLineString, MultiPolygon
import pandas as pd
import geopandas as gpd
from vgrid.generator.settings import geodesic_dggs_to_feature
from vgrid.utils.easedggs.constants import levels_specs, geo_crs, ease_crs
from vgrid.utils.easedggs.dggs.grid_addressing import (
    grid_ids_to_geos,
    geos_to_grid_ids,
    geo_polygon_to_grid_ids,
)
from vgrid.conversion.dggscompact import ease_compact
from vgrid.utils.geometry import (
    check_predicate,
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)


def validate_ease_resolution(resolution):
    """
    Validate that EASE resolution is in the valid range [0..6].

    Args:
        resolution (int): Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [0..6]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 0 or resolution > 6:
        raise ValueError(f"Resolution must be in range [0..6], got {resolution}")

    return resolution


def point2ease(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,  # New parameter for topology preservation
):
    """
    Convert a single point geometry to EASE grid cell.

    Args:
        resolution (int): EASE resolution level [0..6]
        point (shapely.geometry.Point): Point geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for points)
        compact (bool, optional): Enable EASE compact mode (not used for points)
        topology (bool, optional): Enable topology preserving mode - ensures disjoint points have disjoint EASE cells
        include_properties (bool, optional): Whether to include properties in output
        all_points: List of all points for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing EASE cells containing the point
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        
        # Calculate the shortest distance between all points
        shortest_distance = shortest_point_distance(all_points)
        
        # Find resolution where EASE cell size is smaller than shortest distance
        # This ensures disjoint points have disjoint EASE cells
        if shortest_distance > 0:
            for res in range(7):  # EASE resolution range is [0..6]
                if res in levels_specs:
                    cell_width = levels_specs[res]["x_length"]
                    # Use a factor to ensure sufficient separation (cell diagonal is ~1.4x cell width)
                    cell_diagonal = cell_width * 1.4
                    if cell_diagonal < shortest_distance:
                        resolution = res
                        break
            else:
                # If no resolution found, use the highest resolution
                resolution = 6
        else:
            # Single point or no distance, use provided resolution
            pass
    ease_features = []
    latitude = point.y
    longitude = point.x
    ease_cell = geos_to_grid_ids([(longitude, latitude)], level=resolution)
    ease_id = ease_cell["result"]["data"][0]
    level = int(ease_id[1])
    level_spec = levels_specs[level]
    n_row = level_spec["n_row"]
    n_col = level_spec["n_col"]
    geo = grid_ids_to_geos([ease_id])
    center_lon, center_lat = geo["result"]["data"][0]
    cell_min_lat = center_lat - (180 / (2 * n_row))
    cell_max_lat = center_lat + (180 / (2 * n_row))
    cell_min_lon = center_lon - (360 / (2 * n_col))
    cell_max_lon = center_lon + (360 / (2 * n_col))
    cell_polygon = Polygon(
        [
            [cell_min_lon, cell_min_lat],
            [cell_max_lon, cell_min_lat],
            [cell_max_lon, cell_max_lat],
            [cell_min_lon, cell_max_lat],
            [cell_min_lon, cell_min_lat],
        ]
    )
    if cell_polygon:
        num_edges = 4
        ease_feature = geodesic_dggs_to_feature(
            "ease", ease_id, level, cell_polygon, num_edges
        )
        if include_properties and feature_properties:
            ease_feature["properties"].update(feature_properties)
        ease_features.append(ease_feature)
    return ease_features


def polyline2ease(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,  # New parameter for topology preservation
):
    """
    Convert line geometries (LineString, MultiLineString) to EASE grid cells.

    Args:
        resolution (int): EASE resolution level [0..6]
        feature (shapely.geometry.LineString or shapely.geometry.MultiLineString): Line geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for lines)
        compact (bool, optional): Enable EASE compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polylines have disjoint EASE cells
        include_properties (bool, optional): Whether to include properties in output
        all_polylines: List of all polylines for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing EASE cells intersecting the line
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        
        # Calculate the shortest distance between all polylines
        shortest_distance = shortest_polyline_distance(all_polylines)
        
        # Find resolution where EASE cell size is smaller than shortest distance
        # This ensures disjoint polylines have disjoint EASE cells
        if shortest_distance > 0:
            for res in range(7):  # EASE resolution range is [0..6]
                if res in levels_specs:
                    cell_width = levels_specs[res]["x_length"]
                    # Use a factor to ensure sufficient separation (cell diagonal is ~1.4x cell width)
                    cell_diagonal = cell_width * 2.8  # in case there are 2 cells representing the same line segment
                    if cell_diagonal < shortest_distance:
                        resolution = res
                        break
            else:
                # If no resolution found, use the highest resolution
                resolution = 6
        else:
            # Single polyline or no distance, use provided resolution
            pass
    ease_features = []
    if feature.geom_type in ("LineString"):
        polylines = [feature]
    elif feature.geom_type in ("MultiLineString"):
        polylines = list(feature.geoms)
    else:
        return []

    for polyline in polylines:
        poly_bbox = box(*polyline.bounds)
        polygon_bbox_wkt = poly_bbox.wkt
        cells_bbox = geo_polygon_to_grid_ids(
            polygon_bbox_wkt,
            resolution,
            geo_crs,
            ease_crs,
            levels_specs,
            return_centroids=True,
            wkt_geom=True,
        )
        ease_cells = cells_bbox["result"]["data"]
        if compact:
            ease_cells = ease_compact(ease_cells)
        for ease_cell in ease_cells:
            cell_resolution = int(ease_cell[1])
            level_spec = levels_specs[cell_resolution]
            n_row = level_spec["n_row"]
            n_col = level_spec["n_col"]
            geo = grid_ids_to_geos([ease_cell])
            center_lon, center_lat = geo["result"]["data"][0]
            cell_min_lat = center_lat - (180 / (2 * n_row))
            cell_max_lat = center_lat + (180 / (2 * n_row))
            cell_min_lon = center_lon - (360 / (2 * n_col))
            cell_max_lon = center_lon + (360 / (2 * n_col))
            cell_polygon = Polygon(
                [
                    [cell_min_lon, cell_min_lat],
                    [cell_max_lon, cell_min_lat],
                    [cell_max_lon, cell_max_lat],
                    [cell_min_lon, cell_max_lat],
                    [cell_min_lon, cell_min_lat],
                ]
            )
            if cell_polygon.intersects(polyline):
                num_edges = 4
                ease_feature = geodesic_dggs_to_feature(
                    "ease", str(ease_cell), cell_resolution, cell_polygon, num_edges
                )
                if include_properties and feature_properties:
                    ease_feature["properties"].update(feature_properties)
                ease_features.append(ease_feature)
    return ease_features


def polygon2ease(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,  # New parameter for topology preservation
):
    """
    Convert polygon geometries (Polygon, MultiPolygon) to EASE grid cells.

    Args:
        resolution (int): EASE resolution level [0..6]
        feature (shapely.geometry.Polygon or shapely.geometry.MultiPolygon): Polygon geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply ('intersect', 'within', 'centroid_within', 'largest_overlap')
        compact (bool, optional): Enable EASE compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polygons have disjoint EASE cells
        include_properties (bool, optional): Whether to include properties in output
        all_polygons: List of all polygons for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing EASE cells based on predicate
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        
        # Calculate the shortest distance between all polygons
        shortest_distance = shortest_polygon_distance(all_polygons)
        
        # Find resolution where EASE cell size is smaller than shortest distance
        # This ensures disjoint polygons have disjoint EASE cells
        if shortest_distance > 0:
            for res in range(7):  # EASE resolution range is [0..6]
                if res in levels_specs:
                    cell_width = levels_specs[res]["x_length"]
                    # Use a factor to ensure sufficient separation (cell diagonal is ~1.4x cell width)
                    cell_diagonal = cell_width * 2.8
                    if cell_diagonal < shortest_distance:
                        resolution = res
                        break
            else:
                # If no resolution found, use the highest resolution
                resolution = 6
        else:
            # Single polygon or no distance, use provided resolution
            pass
    ease_features = []
    if feature.geom_type in ("Polygon"):
        polygons = [feature]
    elif feature.geom_type in ("MultiPolygon"):
        polygons = list(feature.geoms)
    else:
        return []
    for polygon in polygons:
        poly_bbox = box(*polygon.bounds)
        polygon_bbox_wkt = poly_bbox.wkt
        cells_bbox = geo_polygon_to_grid_ids(
            polygon_bbox_wkt,
            resolution,
            geo_crs,
            ease_crs,
            levels_specs,
            return_centroids=True,
            wkt_geom=True,
        )
        ease_cells = cells_bbox["result"]["data"]
        if compact:
            ease_cells = ease_compact(ease_cells)
        for ease_cell in ease_cells:
            cell_resolution = int(ease_cell[1])
            level_spec = levels_specs[cell_resolution]
            n_row = level_spec["n_row"]
            n_col = level_spec["n_col"]
            geo = grid_ids_to_geos([ease_cell])
            center_lon, center_lat = geo["result"]["data"][0]
            cell_min_lat = center_lat - (180 / (2 * n_row))
            cell_max_lat = center_lat + (180 / (2 * n_row))
            cell_min_lon = center_lon - (360 / (2 * n_col))
            cell_max_lon = center_lon + (360 / (2 * n_col))
            cell_polygon = Polygon(
                [
                    [cell_min_lon, cell_min_lat],
                    [cell_max_lon, cell_min_lat],
                    [cell_max_lon, cell_max_lat],
                    [cell_min_lon, cell_max_lat],
                    [cell_min_lon, cell_min_lat],
                ]
            )
            if check_predicate(cell_polygon, polygon, predicate):
                num_edges = 4
                ease_feature = geodesic_dggs_to_feature(
                    "ease", str(ease_cell), cell_resolution, cell_polygon, num_edges
                )
                if feature_properties:
                    ease_feature["properties"].update(feature_properties)
                ease_features.append(ease_feature)
    return ease_features


# --- Main geometry conversion ---
def geometry2ease(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a list of geometries to EASE grid cells.

    Args:
        geometries (shapely.geometry.BaseGeometry or list): Single geometry or list of geometries
        resolution (int): EASE resolution level [0..6]
        properties_list (list, optional): List of property dictionaries for each geometry
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable EASE compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with EASE grid cells
    """
    resolution = validate_ease_resolution(resolution)
    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    ease_features = []
    for idx, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue
        props = (
            properties_list[idx]
            if properties_list and idx < len(properties_list)
            else {}
        )
        if not include_properties:
            props = {}
        if geom.geom_type == "Point":
            ease_features.extend(
                point2ease(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
            )
        elif geom.geom_type == "MultiPoint":
            for pt in geom.geoms:
                ease_features.extend(
                    point2ease(
                        resolution,
                        pt,
                        props,
                        predicate,
                        compact,
                        topology,
                        include_properties,
                        all_points,  # Pass all points for topology preservation
                    )
                )
        elif geom.geom_type in ("LineString", "MultiLineString"):
            ease_features.extend(
                polyline2ease(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polylines,  # Pass all polylines for topology preservation
                )
            )
        elif geom.geom_type in ("Polygon", "MultiPolygon"):
            ease_features.extend(
                polygon2ease(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polygons=all_polygons,  # Pass all polygons for topology preservation
                )
            )
    return {"type": "FeatureCollection", "features": ease_features}


# --- DataFrame/GeoDataFrame conversion ---
def dataframe2ease(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a pandas DataFrame with geometry column to EASE grid cells.

    Args:
        df (pandas.DataFrame): DataFrame with geometry column
        resolution (int): EASE resolution level [0..6]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable EASE compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with EASE grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2ease(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2ease(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a GeoDataFrame to EASE grid cells.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame to convert
        resolution (int): EASE resolution level [0..6]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable EASE compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with EASE grid cells
    """
    geometries = []
    properties_list = []
    for _, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2ease(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


# --- Main vector2ease function ---
def vector2ease(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """
    Convert vector data to EASE grid cells from various input formats.

    Args:
        data: File path, URL, DataFrame, GeoJSON dict, or Shapely geometry
        resolution (int): EASE resolution [0..6]
        compact (bool): Enable EASE compact mode for polygons (default: False)
        output_format (str): Output format ('geojson', 'gpkg', 'parquet', 'csv', 'shapefile')
        output_path (str): Output file path (optional)
        include_properties (bool): If False, do not include original feature properties. (default: True)
        **kwargs: Additional arguments passed to geopandas read functions
    Returns:
        dict or str: Output in the specified format
    """
    # Process input data directly
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        # GeoDataFrame
        result = geodataframe2ease(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        # DataFrame with geometry column
        result = dataframe2ease(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        # Shapely geometry objects
        result = geometry2ease(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        # GeoJSON dict
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2ease(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        # File path or URL
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2ease(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


# --- Output format conversion ---
def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert EASE result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since EASE uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2ease.gpkg", driver="GPKG")
            return "vector2ease.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2ease.parquet", index=False)
            return "vector2ease.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2ease.shp", driver="ESRI Shapefile")
            return "vector2ease.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


# --- CLI ---
def vector2ease_cli():
    """
    Command-line interface for vector2ease conversion.

    This function provides a command-line interface for converting vector data to EASE grid cells.
    It parses command-line arguments and calls the main vector2ease function.

    Usage:
        python vector2ease.py -i input.shp -r 3 -c -f geojson -o output.geojson

    Arguments:
        -i, --input: Input file path or URL
        -r, --resolution: EASE resolution [0..6]
        -c, --compact: Enable EASE compact mode
        -p, --predicate: Spatial predicate (intersect, within, centroid_within, largest_overlap)
        -t, --topology: Enable topology preserving mode
        -np, --no-props: Do not include original feature properties
        -f, --format: Output format (geojson, gpkg, parquet, csv, shapefile)
        -o, --output: Output file path
    """
    parser = argparse.ArgumentParser(
        description="Convert vector data to EASE grid cells"
    )
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(7),
        metavar="[0-6]",
        help="EASE resolution [0..6] (0=coarsest, 6=finest)",
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable EASE compact mode for polygons",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", help="Enable topology preserving mode"
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_ease_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)
    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2ease{extensions.get(args.format, '')}"
    try:
        vector2ease(
            data,
            args.resolution,
            predicate=args.predicate,
            topology=args.topology,
            compact=args.compact,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2ease_cli()
